{------------------------------------------------------------------------------}
{        This code is copyright (C) 2010-2025 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit FileTools;

interface

uses
  Windows, Messages, UrlMon, ShellAPI, ShLWAPI, ShlObj, {$IF RTLVersion >= 23}System.Win.ComObj{$ELSE}ComObj{$IFEND},
  IsGuids, KnownFolders, ActiveX, Variants, Classes, SysUtils, AnsiStrings, StrTools, WindowTools, Graphics;

const
  MIME_APP = 'application';
  MIME_IMG = 'image';

const
  MIME_SVG  = 'image/svg+xml';
  MIME_WEBP = 'image/webp';
  MIME_PNG  = 'image/png';
  IMG_ALL   = 'image/*';

const
  MIME_TEXT = 'text/plain';
  MIME_HTML = 'text/html';
  MIME_XML  = 'text/xml';
  TEXT_ALL  = 'text/*';

const
  MAX_UNC_PATH = $8000;
  UNC_PREFIX = '\\?\';
  ALL_FILES = '*.*';
  PEEK_SIZE = 256;

const
  FOLDER		= '.folder';
  ASC			= '.asc';
  BAK			= '.bak';
  BMP			= '.bmp';
  CFG			= '.cfg';
  CHM			= '.chm';
  CSS			= '.css';
  DIB			= '.dib';
  DLL			= '.dll';
  EXE			= '.exe';
  FXX			= '.f??';
  GID			= '.gid';
  GIF			= '.gif';
  HLP			= '.hlp';
  HTM			= '.htm';
  ICO			= '.ico';
  JPG			= '.jpg';
  JS			= '.js';
  PNG			= '.png';
  PGP			= '.pgp';
  RTF			= '.rtf';
  SIG			= '.sig';
  SIK			= '.sik';
  SVG			= '.svg';
  TMP			= '.tmp';
  TXT			= '.txt';
  WEBP			= '.webp';
  XML			= '.xml';

const // from LibLoaderApi.h
  // since Windows Vista
  LOAD_IGNORE_CODE_AUTHZ_LEVEL		= $00000010;
  LOAD_LIBRARY_AS_IMAGE_RESOURCE	= $00000020;
  LOAD_LIBRARY_AS_DATAFILE_EXCLUSIVE	= $00000040;
  // since Windows 8
  LOAD_LIBRARY_REQUIRE_SIGNED_TARGET	= $00000080;
  // since Windows Vista
  LOAD_LIBRARY_SEARCH_DLL_LOAD_DIR	= $00000100;
  LOAD_LIBRARY_SEARCH_APPLICATION_DIR	= $00000200;
  LOAD_LIBRARY_SEARCH_USER_DIRS		= $00000400;
  LOAD_LIBRARY_SEARCH_SYSTEM32		= $00000800;
  LOAD_LIBRARY_SEARCH_DEFAULT_DIRS	= $00001000;

type
  TBinaryBitness = (
    bb32,
    bbDos,
    bbWow,
    bbPif,
    bbPosix,
    bbOS2,
    bb64,
    MAX_DWORD
  );

  TFVIParameter = (
    fviFixed,
    fviComments,
    fviCompanyName,
    fviFileDescription,
    fviFileVersion,
    fviInternalName,
    fviLegalCopyright,
    fviLegalTrademarks,
    fviOriginalFilename,
    fviPrivateBuild,
    fviProductName,
    fviProductVersion,
    fviSpecialBuild
  );
  TFVIParameters = Set of TFVIParameter;
  TFileVersionInfo = Record
    QueryFile: String;
    QueryLanguage: String;
    QueryParameters: TFVIParameters;
    Fixed: TVSFixedFileInfo;
    Comments: String;
    CompanyName: String;
    FileDescription: String;
    FileVersion: String;
    InternalName: String;
    LegalCopyright: String;
    LegalTrademarks: String;
    OriginalFilename: String;
    PrivateBuild: String;
    ProductName: String;
    ProductVersion: String;
    SpecialBuild: String;
  end;

  TFileTypeInfos = (
    ftiCommand = 1,
    ftiExecutable,
    ftiFriendlyDocName,
    ftiFriendlyAppName,
    ftiNoOpen,
    ftiShellNewValue,
    ftiDDECommand,
    ftiDDEIfExec,
    ftiDDEApplication,
    ftiDDETopic,
    ftiInfoTip,
    ftiQuickTip,
    ftiTileInfo,
    ftiContentType,
    ftiDefaultIcon,
    ftiShellExtension
  );

  TDriveType = (
    gdtUnknown,
    gdtNoRootDir,
    gdtRemovable,
    gdtFixed,
    gdtRemote,
    gdtCdRom,
    gdtRamDisk
  );

  TDlgFont = Record
    Name: String;
    Size: Integer;
  end;

  THexDigits = 2..8;

type
  TMimeTypeExtensionList = class(TStringList)
  public
    function GetFileTypeByMimeTypeA(const MimeType: String): RawByteString;
    function GetFileTypeByMimeTypeW(const MimeType: String): UnicodeString;
    function GetMimeTypeByExtensionA(const FileNameOrExtension: String): RawByteString;
    function GetMimeTypeByExtensionW(const FileNameOrExtension: String): UnicodeString;
    function GetMimeTypeExtensionA(const FileNameOrExtension, MimeType: String): RawByteString;
    function GetMimeTypeExtensionW(const FileNameOrExtension, MimeType: String): UnicodeString;
  end;

var
  MimeTypeExtensionList: TMimeTypeExtensionList = NIL;

function LastDelimiter(const FilePath: RawByteString; ExtDotPos: Longbool = FALSE): Integer; overload;
function LastDelimiter(const FilePath: UnicodeString; ExtDotPos: Longbool = FALSE): Integer; overload;
function ExtractFileDrive(const FileName: RawByteString; Unquote: Longbool = TRUE): RawByteString; overload;
function ExtractFileDrive(const FileName: UnicodeString; Unquote: Longbool = TRUE): UnicodeString; overload;
function ExtractFilePath(const FileName: RawByteString; Unquote: Longbool = TRUE): RawByteString; overload;
function ExtractFilePath(const FileName: UnicodeString; Unquote: Longbool = TRUE): UnicodeString; overload;
function ExtractFileName(const FileName: RawByteString; Unquote: Longbool = TRUE): RawByteString; overload;
function ExtractFileName(const FileName: UnicodeString; Unquote: Longbool = TRUE): UnicodeString; overload;
function ExtractBaseName(const FileName: RawByteString; Unquote: Longbool = TRUE): RawByteString; overload;
function ExtractBaseName(const FileName: UnicodeString; Unquote: Longbool = TRUE): UnicodeString; overload;
function ExtractFileExt(const FileName: RawByteString; Unquote: Longbool = TRUE): RawByteString; overload;
function ExtractFileExt(const FileName: UnicodeString; Unquote: Longbool = TRUE): UnicodeString; overload;
function ChangeFileExt(const FileName, NewExtension: RawByteString): RawByteString; overload;
function ChangeFileExt(const FileName, NewExtension: UnicodeString): UnicodeString; overload;
function HasFileExt(const FileName, TestExtension: RawByteString): Longbool; overload;
function HasFileExt(const FileName, TestExtension: UnicodeString): Longbool; overload;
function IsFileExt(const FileName, TestExtension: RawByteString): Longbool; overload;
function IsFileExt(const FileName, TestExtension: UnicodeString): Longbool; overload;
function SameFileExt(const Ext1, Ext2: RawByteString): Longbool; overload;
function SameFileExt(const Ext1, Ext2: UnicodeString): Longbool; overload;
function FindFileExt(const FileName: RawByteString): Integer; overload;
function FindFileExt(const FileName: UnicodeString): Integer; overload;
function SameTextFiles(const Name1, Name2: RawByteString; SampleSize: Cardinal): Longbool; overload;
function SameTextFiles(const Name1, Name2: UnicodeString; SampleSize: Cardinal): Longbool; overload;
function SameFileDriveTypes(const Name1, Name2: RawByteString): TDriveType; overload;
function SameFileDriveTypes(const Name1, Name2: UnicodeString): TDriveType; overload;
function SameFileNames(const Name1, Name2: RawByteString): Longbool; overload;
function SameFileNames(const Name1, Name2: UnicodeString): Longbool; overload;
function IsLegalFileName(const FileName: RawByteString): Longbool; overload;
function IsLegalFileName(const FileName: UnicodeString): Longbool; overload;
function GetLegalFileName(const FileName: RawByteString): RawByteString; overload;
function GetLegalFileName(const FileName: UnicodeString): UnicodeString; overload;
function GetLegalFileExt(const FileName: RawByteString): RawByteString; overload;
function GetLegalFileExt(const FileName: UnicodeString): UnicodeString; overload;
procedure AddBackslash(var Path: RawByteString; PathLen: Integer); overload;
procedure AddBackslash(var Path: UnicodeString; PathLen: Integer); overload;
function AddBackslash(const Path: RawByteString): RawByteString; overload;
function AddBackslash(const Path: UnicodeString): UnicodeString; overload;
procedure RemoveBackslash(var Path: String; PathLen: Integer); overload;
function RemoveBackslash(const Path: String): String; overload;
function GetAppNameFromCommandLine(const CommandLine: String): String;
function GetAppPathFromCommandLine(const CommandLine: String): String;
function ReplaceFileDrive(const Path, Drive: String): String;
function ReplaceFileName(const Path, Name: String): String;
function GetQuotedPath(const Path: String): String;
function GetShortPath(const Path: String): String;
function GetLongPath(const Path: String): String;
function GetParentDir(const Dir: String): String;
function CreateSubDir(const Path: String): String;
function RemoveSubDirs(const FullPath, BasePath: String): String;
function SetCurDir(const Name: String): Longbool;
function GetCurDir: String;
function GetWinDir: String;
function GetSysDir: String;
{$IFDEF UNICODE}
function GetTmpDir: UnicodeString;
{$ELSE}
function GetTmpDir: RawByteString;
{$ENDIF}
function GetTmpDirA: RawByteString;
function GetTmpDirW: UnicodeString;
function GetModuleName(ModuleHandle: THandle): String;
function GetModulePath(ModuleHandle: THandle): String;
function GetModuleOrAppPath(const FileName: String; const AppPath: String): String;
function GetUniqueFileName(const FileExt: String; Hashed: Longbool = TRUE): String;
function GetDriveType(const Name: String): TDriveType;
function GetUNCFileName(const FileName: String): String;
function GetFileNameByHandle(FileHandle: THandle): String;
function GetLocalizedPath(const Path: String): String;
function GetOrLoadLibrary(const Name: String): THandle;
function LoadLibraryEx(const Name: String; Flags: Cardinal): THandle;
function LoadMuiLibrary(const LibInfo: Variant; Localize: Longbool): THandle;
function MakeErrorFileExtension(Code: Cardinal; Digits: THexDigits = 4): String;
function SetErrorFileExtension(const FileName: String; Code: Cardinal; Digits: THexDigits = 4): String;
function NoValidFile(const FileName: RawByteString): Longbool; overload;
function NoValidFile(const FileName: UnicodeString): Longbool; overload;
function ValidFile(const FileName: RawByteString): Longbool; overload;
function ValidFile(const FileName: UnicodeString): Longbool; overload;
function IsExecutable(const FileName: RawByteString): Longbool; overload;
function IsExecutable(const FileName: UnicodeString): Longbool; overload;
function IsBinaryBitness(const FileName: RawByteString): TBinaryBitness; overload;
function IsBinaryBitness(const FileName: UnicodeString): TBinaryBitness; overload;
function GetSystemFolderPath(FolderID: Integer; Localize: Longbool = FALSE): String; overload;
function GetSystemFolderPath(FolderID: PItemIDList; Localize: Longbool = FALSE): String; overload;
function GetSystemFolderPath(const KnownFolderID: TIID; Localize: Longbool = FALSE): String; overload;
function CreateTempFileA(Path: RawByteString = E; Prefix: RawByteString = 'TMP'; Keep: Longbool = FALSE): RawByteString; overload;
function CreateTempFileW(Path: UnicodeString = E; Prefix: UnicodeString = 'TMP'; Keep: Longbool = FALSE): UnicodeString; overload;
function CreateTempFileByName(const FileName: RawByteString): RawByteString; overload;
function CreateTempFileByName(const FileName: UnicodeString): UnicodeString; overload;
function CreateLinkFile(const Link, Target, Commands, Icon: RawByteString): Longbool; overload;
function CreateLinkFile(const Link, Target, Commands, Icon: UnicodeString): Longbool; overload;
function CreateUrlShortcut(const FileName, Url: RawByteString): Longbool; overload;
function CreateUrlShortcut(const FileName, Url: UnicodeString): Longbool; overload;
function GetLinkTarget(const LinkFile: RawByteString): RawByteString; overload;
function GetLinkTarget(const LinkFile: UnicodeString): UnicodeString; overload;
function ReadWriteAccess(const FileOrFolder: RawByteString): Longbool; overload;
function ReadWriteAccess(const FileOrFolder: UnicodeString): Longbool; overload;
function ReadFromConsole: RawByteString;
procedure WriteToConsole(const Text: RawByteString); overload;
procedure WriteToConsole(const Text: UnicodeString); overload;
function CopyFile(const Source, Dest: RawByteString; FailIfExists: Longbool): Longbool; overload;
function CopyFile(const Source, Dest: UnicodeString; FailIfExists: Longbool): Longbool; overload;
function MoveFile(const Source, Dest: RawByteString; Overwrite: Longbool): Longbool; overload;
function MoveFile(const Source, Dest: UnicodeString; Overwrite: Longbool): Longbool; overload;
function IsFileReadOnly(const FileName: RawByteString): Longbool; overload;
function IsFileReadOnly(const FileName: UnicodeString): Longbool; overload;
function SetFileReadOnly(const FileName: RawByteString; Value: Boolean): Boolean; overload;
function SetFileReadOnly(const FileName: UnicodeString; Value: Boolean): Boolean; overload;
function RenameWriteProtect(const Old, New: RawByteString; Overwrite, Protect: Boolean): Longbool; overload;
function RenameWriteProtect(const Old, New: UnicodeString; Overwrite, Protect: Boolean): Longbool; overload;
function AppendTextFile(const Name: RawByteString; const Text: RawByteString; AsLine: Longbool = false): Longbool; overload;
function AppendTextFile(const Name: UnicodeString; const Text: RawByteString; AsLine: Longbool = false): Longbool; overload;
function AppendTextFile(const Name: UnicodeString; const Text: UnicodeString; AsLine: Longbool = false): Longbool; overload;
function WriteTextFile(const Name: RawByteString; const Text: RawByteString): Longbool; overload;
function WriteTextFile(const Name: UnicodeString; const Text: RawByteString): Longbool; overload;
function WriteTextFile(const Name: UnicodeString; const Text: UnicodeString): Longbool; overload;
function WriteTempFile(const Name, Prefix: RawByteString; const Text: RawByteString): RawByteString; overload;
function WriteTempFile(const Name, Prefix: UnicodeString; const Text: RawByteString): UnicodeString; overload;
function WriteTempFile(const Name, Prefix: UnicodeString; const Text: UnicodeString): UnicodeString; overload;
function ShowTextFile(const Name: RawByteString; const Text: RawByteString): Longbool; overload;
function ShowTextFile(const Name: UnicodeString; const Text: RawByteString): Longbool; overload;
function ShowTextFile(const Name: UnicodeString; const Text: UnicodeString): Longbool; overload;
function ReadTextFile(const Name: RawByteString; Bytes: Integer = 0; CheckBom: Longbool = FALSE): RawByteString; overload;
function ReadTextFile(const Name: UnicodeString; Bytes: Integer = 0; CheckBom: Longbool = FALSE): RawByteString; overload;
function ReadTextFileAsUtf8(const Name: String; const Charset: RawByteString; IncludeBom: Longbool = FALSE): Utf8String; inline;
function ReadTextFileAsUnicode(const Name: String; Charset: RawByteString; Utf8: Longbool = FALSE; IncludeBom: Longbool = FALSE): String;
function CreateOrOpenFile(const Name: RawByteString): THandle; overload;
function CreateOrOpenFile(const Name: UnicodeString): THandle; overload;
function CreateWriteFile(const Name: RawByteString): THandle; overload;
function CreateWriteFile(const Name: UnicodeString): THandle; overload;
function CreateEmptyFile(const Name: RawByteString): RawByteString; overload;
function CreateEmptyFile(const Name: UnicodeString): UnicodeString; overload;
function TestWriteFile(const Name: RawByteString): Longbool; overload;
function TestWriteFile(const Name: UnicodeString): Longbool; overload;
function OpenReadFile(const Name: RawByteString): THandle; overload;
function OpenReadFile(const Name: UnicodeString): THandle; overload;
function CloseFile(const Handle: THandle): Longbool;
function FileAge(const Name: RawByteString): TDateTime; overload;
function FileAge(const Name: UnicodeString): TDateTime; overload;
function FileSize(const Name: RawByteString): Int64; overload;
function FileSize(const Name: UnicodeString): Int64; overload;
function GetFileTypeInfo(const FileExtension: RawByteString; fti: TFileTypeInfos): RawByteString; overload;
function GetFileTypeInfo(const FileExtension: UnicodeString; fti: TFileTypeInfos): UnicodeString; overload;
function UpdateFileAccessTime(const Name: RawByteString; Time: TDateTime = 0): Longbool; overload;
function UpdateFileAccessTime(const Name: UnicodeString; Time: TDateTime = 0): Longbool; overload;
function FileVersionInfo(var FileVersionInfo: TFileVersionInfo): Longbool;
function LoadMimeTypeExtensionList(var MTEList: TMimeTypeExtensionList): Cardinal;
function GetMimeTypeExtension(const MimeType: String): String;
function ExtractMimeTypeExtension(const FileName: String): String;
function ValidateFileExtension(const FileNameOrExtension: String): String;
function IsFileExtSvgXml(const FileNameExt1, FileNameExt2: String): Longbool;
function GetDataMimeType(Data: Pointer; Size: Cardinal; const MimeProposed: String = E): String;
function GetFileMimeType(const FileName: String; const DefaultType: String = E; const FileData: RawByteString = E): String;
function IsApplicationType(const FileName: String; const DefaultType: String = E; const FileData: RawByteString = E): Boolean;
function IsImageType(const FileName: String; const DefaultType: String = E; const FileData: RawByteString = E): Boolean;
function IsHyperTextType(const FileName: String): Boolean;
function IsSvgType(const FileName: String): Boolean;
function FindFile(const FileMask: AnsiString; const ExcludeExt: AnsiString = E): AnsiString; overload;
function FindFile(const FileMask: UnicodeString; const ExcludeExt: UnicodeString = E): UnicodeString; overload;
function FindFirst(const FileMask: AnsiString; var FindData: TWin32FindDataA): THandle; overload;
function FindFirst(const FileMask: UnicodeString; var FindData: TWin32FindDataW): THandle; overload;
function FindNext(Search: THandle; var FindData: TWin32FindDataA): Longbool; overload;
function FindNext(Search: THandle; var FindData: TWin32FindDataW): Longbool; overload;
function FindClose(Search: THandle): Longbool;
function DeleteFiles(const FileMask: AnsiString): Longbool; overload;
function DeleteFiles(const FileMask: UnicodeString): Longbool; overload;
function DeleteFile(const Name: RawByteString; IncludeReadOnly: Longbool = FALSE): Longbool; overload;
function DeleteFile(const Name: UnicodeString; IncludeReadOnly: Longbool = FALSE): Longbool; overload;
function StartsDSorBS(const Name: RawByteString): Longbool; overload;
function StartsDSorBS(const Name: UnicodeString): Longbool; overload;
function FileExists(const Name: RawByteString; IncludeRemote: Longbool = FALSE): Longbool; overload;
function FileExists(const Name: UnicodeString; IncludeRemote: Longbool = FALSE): Longbool; overload;
function DirExists(const Name: RawByteString; LastBackslashPos: Integer = 0): Longbool; overload;
function DirExists(const Name: UnicodeString; LastBackslashPos: Integer = 0): Longbool; overload;
function FindDir(Owner: THandle; const SearchPath, Prompt: String; AllowCreation: Longbool = FALSE; EnableEdit: Longbool = FALSE): String;
function FindExecutable(const FileExtension: String): String;
function ReloadModule(const FileName: String): THandle;
function LoadResModule(const FileName: String): THandle;
function ReloadAsResModule(Module: THandle): THandle;
function DialogItemInt(Dialog: THandle; ItemID: Cardinal): Integer;
function DialogItemString(Dialog: THandle; ItemID: Cardinal): String;
function DialogFontInfo(Module: THandle; DialogID: PChar): TDlgFont;
function DialogCaptionString(Module: THandle; DialogID: PChar): String;
function DialogControlString(Module: THandle; DialogID: PChar; ControlID: Cardinal): String;
function MenuItemString(Module: THandle; MenuID: PChar; ItemID: Cardinal): String;
function LoadHtmlString(Module: THandle; const Name: String): RawByteString;

implementation

uses
  Processes, UTF8;

const	// WinNT
  FILE_READ_ATTRIBUTES		= $0080;	// all

  FILE_WRITE_ATTRIBUTES		= $0100;	// all

  FILE_ALL_ACCESS		= (STANDARD_RIGHTS_REQUIRED or SYNCHRONIZE or $01FF);

// from winbase.h
const
  SCS_32BIT_BINARY	= 0;
  SCS_DOS_BINARY	= 1;
  SCS_WOW_BINARY	= 2;
  SCS_PIF_BINARY	= 3;
  SCS_POSIX_BINARY	= 4;
  SCS_OS216_BINARY	= 5;
  SCS_64BIT_BINARY	= 6;

type
  PBasicInfo = ^TBasicInfo;
  TBasicInfo = Record
    CreationTime: LargeInt;
    LastAccessTime: LargeInt;
    LastWriteTime: LargeInt;
    ChangeTime: LargeInt;
    FileAttributes: DWord;
  end;

  PStandardInfo = ^TStandardInfo;
  TStandardInfo = Record
    AllocationSize: LargeInt;
    EndOfFile: LargeInt;
    NumberOfLinks: DWord;
    DeletePending: Boolean;
    Directory: Boolean;
  end;

  PNameInfo = ^TNameInfo;
  TNameInfo = Record
    FileNameLength: DWord;
    FileName: Array[0..0] of WChar;
  end;

  PRenameInfo = ^TRenameInfo;
  TRenameInfo = Record
    ReplaceIfExists: Boolean;
    RootDirectory: THandle;
    FileNameLength: DWord;
    FileName: Array[0..0] of WChar;
  end;

  PAllocationInfo = ^TAllocationInfo;
  TAllocationInfo = Record
    AllocationSize: LargeInt;
  end;

  PEndOftInfo = ^TEndOftInfo;
  TEndOftInfo = Record
    EndOfFile: LargeInt;
  end;

  PStreamInfo = ^TStreamInfo;
  TStreamInfo = Record
    NextEntryOffset: DWord;
    StreamNameLength: DWord;
    StreamSize: LargeInt;
    StreamAllocationSize: LargeInt;
    StreamName: Array[0..0] of WChar;
  end;

  PCompressionInfo = ^TCompressionInfo;
  TCompressionInfo = Record
    CompressedFileSize: LargeInt;
    CompressionFormat: Word;
    CompressionUnitShift: UChar;
    ChunkShift: UChar;
    ClusterShift: UChar;
    Reserved: Array[0..2] of UChar;
  end;

  PAttributeTagInfo = ^TAttributeTagInfo;
  TAttributeTagInfo = Record
    FileAttributes: DWord;
    ReparseTag: DWord;
  end;

  PDispositionInfo = ^TDispositionInfo;
  TDispositionInfo = Record
    DeleteFile: Boolean;
  end;

 // from minwinbase.h
type
  TFileInfoByHandleClass = (
    FileBasicInfo,
    FileStandardInfo,
    FileNameInfo,
    FileRenameInfo,
    FileDispositionInfo,
    FileAllocationInfo,
    FileEndOfFileInfo,
    FileStreamInfo,
    FileCompressionInfo,
    FileAttributeTagInfo,
    FileIdBothDirectoryInfo,
    FileIdBothDirectoryRestartInfo,
    FileIoPriorityHintInfo,
    FileRemoteProtocolInfo,
    FileFullDirectoryInfo,
    FileFullDirectoryRestartInfo,
    FileStorageInfo,
    FileAlignmentInfo,
    FileIdInfo,
    FileIdExtdDirectoryInfo,
    FileIdExtdDirectoryRestartInfo,
    FileDispositionInfoEx,
    FileRenameInfoEx,
    FileCaseSensitiveInfo,
    FileNormalizedNameInfo,
    MaximumFileInfoByTHandleClass
  );

// from fileapi.h
function GetFileSizeEx(hFile: THandle; var lpFileSize: LARGE_INTEGER): Bool; stdcall;
		       external kernel32 name 'GetFileSizeEx';
// from winbase.h
function GetFileInformationByHandleEx(hFile: THandle; FileInfoByHandleClass: TFileInfoByHandleClass;
				      lpFileInformation: Pointer; dwBufferSize: DWord): Bool; stdcall;
				      external kernel32 name 'GetFileInformationByHandleEx';


// my stuff
const
  ILLEGAL_FN_CHARS = ['"', '*', '/', ':', '<', '>', '?', '\', '|'];

function TMimeTypeExtensionList.GetFileTypeByMimeTypeA(const MimeType: String): RawByteString;
begin
  Result := RAW(Values[MimeType]);
end;

function TMimeTypeExtensionList.GetFileTypeByMimeTypeW(const MimeType: String): UnicodeString;
begin
  Result := Values[MimeType];
end;

function TMimeTypeExtensionList.GetMimeTypeByExtensionA(const FileNameOrExtension: String): RawByteString;
var
  sFileExt: String;
  iExt, ilen, iIndex: Integer;
  pPos: PChar;
begin
  sFileExt := FileTools.ExtractFileExt(FileNameOrExtension);
  iExt := Length(sFileExt);
  if iExt > 1 then for iIndex := 0 to pred(Count) do begin
    ilen := Length(Strings[iIndex]) - iExt;
    pPos := Pointer(Strings[iIndex]);
    while (pPos^ <> NUL) and (pPos^ <> SL) do inc(pPos);
    if (StrTools.StrLIComp(PChar(Strings[iIndex]) + iLen, PChar(sFileExt), iExt) = 0)
    or (StrTools.StrLIComp(pPos + 1, PChar(sFileExt) + 1, pred(iExt)) = 0) then begin
      Result := RAW(Names[iIndex]);
      Exit;
    end;
  end;
  Result := E;
end;

function TMimeTypeExtensionList.GetMimeTypeByExtensionW(const FileNameOrExtension: String): UnicodeString;
var
  sFileExt: String;
  iExt, ilen, iIndex: Integer;
  pPos: PChar;
begin
  sFileExt := FileTools.ExtractFileExt(FileNameOrExtension);
  iExt := Length(sFileExt);
  if iExt > 1 then for iIndex := 0 to pred(Count) do begin
    ilen := Length(Strings[iIndex]) - iExt;
    pPos := Pointer(Strings[iIndex]);
    while (pPos^ <> NUL) and (pPos^ <> SL) do inc(pPos);
    if (StrTools.StrLIComp(PChar(Strings[iIndex]) + iLen, PChar(sFileExt), iExt) = 0)
    or (StrTools.StrLIComp(pPos + 1, PChar(sFileExt) + 1, pred(iExt)) = 0) then begin
      Result := Names[iIndex];
      Exit;
    end;
  end;
  Result := E;
end;

function TMimeTypeExtensionList.GetMimeTypeExtensionA(const FileNameOrExtension, MimeType: String): RawByteString;
var
  sMimeExt, sFileExt: String;
begin
  sMimeExt := Values[MimeType];
  sFileExt := GetLegalFileExt(FileNameOrExtension);
  if (sMimeExt <> E) and not SameFileExt(sFileExt, sMimeExt) then
    Result := StrTools.LowerCase(RAW(sMimeExt))
  else Result := StrTools.LowerCase(RAW(sFileExt));
end;

function TMimeTypeExtensionList.GetMimeTypeExtensionW(const FileNameOrExtension, MimeType: String): UnicodeString;
var
  sMimeExt, sFileExt: String;
begin
  sMimeExt := Values[MimeType];
  sFileExt := GetLegalFileExt(FileNameOrExtension);
  if (sMimeExt <> E) and not SameFileExt(sFileExt, sMimeExt) then
    Result := StrTools.LowerCase(sMimeExt)
  else Result := StrTools.LowerCase(sFileExt);
end;

function LastDelimiter(const FilePath: RawByteString; ExtDotPos: Longbool = FALSE): Integer;
begin
  if ExtDotPos then
    Result := AnsiStrings.LastDelimiter(DOT + PathDelim + DriveDelim, FilePath)
  else Result := AnsiStrings.LastDelimiter(PathDelim + DriveDelim, FilePath);
end;

function LastDelimiter(const FilePath: UnicodeString; ExtDotPos: Longbool = FALSE): Integer;
begin
  if ExtDotPos then
    Result := SysUtils.LastDelimiter(DOT + PathDelim + DriveDelim, FilePath)
  else Result := SysUtils.LastDelimiter(PathDelim + DriveDelim, FilePath);
end;

function ExtractFileDrive(const FileName: RawByteString; Unquote: Longbool = TRUE): RawByteString;
begin
  if FileName <> E then begin
    if Unquote then
      Result := AnsiStrings.ExtractFileDrive(StrTools.Unquote(FileName))
    else Result := AnsiStrings.ExtractFileDrive(FileName);
    Exit;
  end;
  Result := E;
end;

function ExtractFileDrive(const FileName: UnicodeString; Unquote: Longbool = TRUE): UnicodeString;
begin
  if FileName <> E then begin
    if Unquote then
      Result := SysUtils.ExtractFileDrive(StrTools.Unquote(FileName))
    else Result := SysUtils.ExtractFileDrive(FileName);
    Exit;
  end;
  Result := E;
end;

function ExtractFilePath(const FileName: RawByteString; Unquote: Longbool = TRUE): RawByteString;
begin
  if FileName <> E then begin
    if Unquote then
      Result := AnsiStrings.ExtractFilePath(StrTools.Unquote(FileName))
    else Result := AnsiStrings.ExtractFilePath(FileName);
    Exit;
  end;
  Result := E;
end;

function ExtractFilePath(const FileName: UnicodeString; Unquote: Longbool = TRUE): UnicodeString;
begin
  if FileName <> E then begin
    if Unquote then
      Result := SysUtils.ExtractFilePath(StrTools.Unquote(FileName))
    else Result := SysUtils.ExtractFilePath(FileName);
    Exit;
  end;
  Result := E;
end;

function ExtractFileName(const FileName: RawByteString; Unquote: Longbool = TRUE): RawByteString;
begin
  if FileName <> E then begin
    if Unquote then
      Result := AnsiStrings.ExtractFileName(StrTools.Unquote(FileName))
    else Result := AnsiStrings.ExtractFileName(FileName);
    Exit;
  end;
  Result := E;
end;

function ExtractFileName(const FileName: UnicodeString; Unquote: Longbool = TRUE): UnicodeString;
begin
  if FileName <> E then begin
    if Unquote then
      Result := SysUtils.ExtractFileName(StrTools.Unquote(FileName))
    else Result := SysUtils.ExtractFileName(FileName);
    Exit;
  end;
  Result := E;
end;

function ExtractBaseName(const FileName: RawByteString; Unquote: Longbool = TRUE): RawByteString;
var
  iPos: Integer;
begin
  if FileName <> E then begin
    Result := FileTools.ExtractFileName(StrTools.Unquote(FileName));
    iPos := LastCharPos(DOT, Result);
    if iPos > 0 then Delete(Result, iPos, Length(Result));
    Exit;
  end;
  Result := E;
end;

function ExtractBaseName(const FileName: UnicodeString; Unquote: Longbool = TRUE): UnicodeString;
var
  iPos: Integer;
begin
  if FileName <> E then begin
    Result := FileTools.ExtractFileName(StrTools.Unquote(FileName));
    iPos := LastCharPos(DOT, Result);
    if iPos > 0 then Delete(Result, iPos, Length(Result));
    Exit;
  end;
  Result := E;
end;

function ExtractFileExt(const FileName: RawByteString; Unquote: Longbool = TRUE): RawByteString;

  function GetLastDotPos(const FileName: RawByteString): Integer;
  var
    pBeg, pEnd: PAnsiChar;
  begin
    pBeg := Pointer(FileName);
    pEnd := pBeg + pred(Length(FileName));
    while (pEnd >= pBeg) and not StrTools.CharInSet(pEnd^, [DOT, SL, CL, BS]) do dec(pEnd);
    Result := pEnd - pBeg;
  end;

var
  iDotPos: Integer;
begin
  if FileName <> E then begin
    if not Unquote then
      Result := FileName
    else Result := StrTools.Unquote(FileName);
    iDotPos := GetLastDotPos(Result);
    if iDotPos > -1 then begin
      Delete(Result, 1, iDotPos);
      Exit;
    end;
  end;
  Result := E;
end;

function ExtractFileExt(const FileName: UnicodeString; Unquote: Longbool = TRUE): UnicodeString;

  function GetLastDotPos(const FileName: UnicodeString): Integer;
  var
    pBeg, pEnd: PWideChar;
  begin
    pBeg := Pointer(FileName);
    pEnd := pBeg + pred(Length(FileName));
    while (pEnd >= pBeg) and not CharInSet(pEnd^, [DOT, SL, CL, BS]) do dec(pEnd);
    Result := pEnd - pBeg;
  end;

var
  iDotPos: Integer;
begin
  if FileName <> E then begin
    if not Unquote then
      Result := FileName
    else Result := StrTools.Unquote(FileName);
    iDotPos := GetLastDotPos(Result);
    if iDotPos > -1 then begin
      Delete(Result, 1, iDotPos);
      Exit;
    end;
  end;
  Result := E;
end;

function ChangeFileExt(const FileName, NewExtension: RawByteString): RawByteString;
begin
  Result := AnsiStrings.ChangeFileExt(FileName, NewExtension);
end;

function ChangeFileExt(const FileName, NewExtension: UnicodeString): UnicodeString;
begin
  Result := SysUtils.ChangeFileExt(FileName, NewExtension);
end;

function HasFileExt(const FileName, TestExtension: RawByteString): Longbool; inline;
begin
  Result := IsFileExt(FileName, TestExtension);
end;

function HasFileExt(const FileName, TestExtension: UnicodeString): Longbool; inline;
begin
  Result := IsFileExt(FileName ,TestExtension);
end;

function IsFileExt(const FileName, TestExtension: RawByteString): Longbool;
var
  iPos: Integer;
begin
  iPos := FindFileExt(FileName);
  if iPos > 0 then begin
    Result := TextAtPos(TestExtension, FileName, iPos) and (iPos + Length(TestExtension) = succ(Length(FileName)));
    Exit;
  end;
  Result := FALSE;
end;

function IsFileExt(const FileName, TestExtension: UnicodeString): Longbool;
var
  iPos: Integer;
begin
  iPos := FindFileExt(FileName);
  if iPos > 0 then begin
    Result := TextAtPos(TestExtension, FileName, iPos) and (iPos + Length(TestExtension) = succ(Length(FileName)));
    Exit;
  end;
  Result := FALSE;
end;

function SameFileExt(const Ext1, Ext2: RawByteString): Longbool;
begin
  Result := IsFileExt(Ext1, Ext2);
end;

function SameFileExt(const Ext1, Ext2: UnicodeString): Longbool;
begin
  Result := IsFileExt(Ext1, Ext2);
end;

function FindFileExt(const FileName: RawByteString): Integer;
begin
  Result := LastDelimiter(FileName, TRUE);
  if (Result = 0) or (FileName[Result] <> DOT) then Result := 0;
end;

function FindFileExt(const FileName: UnicodeString): Integer;
begin
  Result := LastDelimiter(FileName, TRUE);
  if (Result = 0) or (FileName[Result] <> DOT) then Result := 0;
end;

function SameTextFiles(const Name1, Name2: RawByteString; SampleSize: Cardinal): Longbool;
var
  fs1, fs2: TFileStream;
  dw1, dw2: DWord;
  rs1, rs2: RawByteString;
begin
  fs1 := NIL;
  fs2 := NIL;
  Result := false;
  try
    fs1 := TFileStream.Create(String(Name1), fmOpenRead or fmShareDenyWrite);
    if fs1 = NIL then Exit;
    fs2 := TFileStream.Create(String(Name2), fmOpenRead or fmShareDenyWrite);
    if fs2 = NIL then Exit;
    if fs1.Size <> fs2.Size then Exit;
    SetString(rs1, NIL, SampleSize);
    SetString(rs2, NIL, SampleSize);
    dw1 := fs1.Read(Pointer(rs1)^, SampleSize);
    dw2 := fs2.Read(Pointer(rs2)^, SampleSize);
    if dw1 <> dw2 then Exit;
    if StrTools.StrLIComp(PAnsiChar(rs1), PAnsiChar(rs2), SampleSize) <> 0 then Exit;
    fs1.Seek(-2048, soEnd);
    fs2.Seek(-2048, soEnd);
    dw1 := fs1.Read(Pointer(rs1)^, SampleSize);
    dw2 := fs2.Read(Pointer(rs2)^, SampleSize);
    if dw1 <> dw2 then Exit;
    Result := (StrTools.StrLIComp(PAnsiChar(rs1), PAnsiChar(rs2), SampleSize) = 0);
  finally
    if fs1 <> NIL then fs1.Free;
    if fs2 <> NIL then fs2.Free;
  end;
end;

function SameTextFiles(const Name1, Name2: UnicodeString; SampleSize: Cardinal): Longbool;
var
  fs1, fs2: TFileStream;
  dw1, dw2: DWord;
  rs1, rs2: RawByteString;
begin
  fs1 := NIL;
  fs2 := NIL;
  Result := false;
  try
    fs1 := TFileStream.Create(Name1, fmOpenRead or fmShareDenyWrite);
    if fs1 = NIL then Exit;
    fs2 := TFileStream.Create(Name2, fmOpenRead or fmShareDenyWrite);
    if fs2 = NIL then Exit;
    if fs1.Size <> fs2.Size then Exit;
    SetString(rs1, NIL, SampleSize);
    SetString(rs2, NIL, SampleSize);
    dw1 := fs1.Read(Pointer(rs1)^, SampleSize);
    dw2 := fs2.Read(Pointer(rs2)^, SampleSize);
    if dw1 <> dw2 then Exit;
    if StrTools.StrLIComp(PAnsiChar(rs1), PAnsiChar(rs2), SampleSize) <> 0 then Exit;
    fs1.Seek(-2048, soEnd);
    fs2.Seek(-2048, soEnd);
    dw1 := fs1.Read(Pointer(rs1)^, SampleSize);
    dw2 := fs2.Read(Pointer(rs2)^, SampleSize);
    if dw1 <> dw2 then Exit;
    Result := (StrTools.StrLIComp(PAnsiChar(rs1), PAnsiChar(rs2), SampleSize) = 0);
  finally
    if fs1 <> NIL then fs1.Free;
    if fs2 <> NIL then fs2.Free;
  end;
end;

function SameFileDriveTypes(const Name1, Name2: RawByteString): TDriveType;
var
  sExt1, sExt2: RawByteString;
begin
  sExt1 := AnsiStrings.ExtractFileDrive(Name1);
  sExt2 := AnsiStrings.ExtractFileDrive(Name2);
  if StrTools.SameText(sExt1, sExt2) then begin
    Result := TDriveType(Windows.GetDriveTypeA(PAnsiChar(sExt1 + BS)));
    if TDriveType(Windows.GetDriveTypeA(PAnsiChar(sExt2 + BS))) <> Result then Result := gdtUnknown;
  end
  else Result := gdtUnknown;
end;

function SameFileDriveTypes(const Name1, Name2: UnicodeString): TDriveType;
var
  sExt1, sExt2: UnicodeString;
begin
  sExt1 := SysUtils.ExtractFileDrive(Name1);
  sExt2 := SysUtils.ExtractFileDrive(Name2);
  if StrTools.SameText(sExt1, sExt2) then begin
    Result := TDriveType(Windows.GetDriveTypeW(PWideChar(sExt1 + BS)));
    if TDriveType(Windows.GetDriveTypeW(PWideChar(sExt2 + BS))) <> Result then Result := gdtUnknown;
  end
  else Result := gdtUnknown;
end;

function SameFileNames(const Name1, Name2: RawByteString): Longbool;
const
  DELIMS = [SL, CL, BS];
var
  pBeg1, pBeg2, pEnd1, pEnd2: PAnsiChar;
begin
  Result := FALSE;
  if (Name1 <> E) and (Name2 <> E) then begin
    pBeg1:= Pointer(Name1);
    pBeg2:= Pointer(Name2);
    pEnd1 := pBeg1;
    pEnd2 := pBeg2;
    inc(pEnd1, Length(Name1));
    inc(pEnd2, Length(Name2));
    while (pEnd1 > pBeg1) and not StrTools.CharInSet(pEnd1^, DELIMS) do dec(pEnd1);
    while (pEnd2 > pBeg2) and not StrTools.CharInSet(pEnd2^, DELIMS) do dec(pEnd2);
    if pEnd1 > pBeg1 then inc(pEnd1);
    if pEnd2 > pBeg2 then inc(pEnd2);
    Result := (StrTools.StrIComp(pEnd1, pEnd2) = 0);
  end;
end;

function SameFileNames(const Name1, Name2: UnicodeString): Longbool;
const
  DELIMS = [SL, CL, BS];
var
  pBeg1, pBeg2, pEnd1, pEnd2: PWideChar;
begin
  Result := FALSE;
  if (Name1 <> E) and (Name2 <> E) then begin
    pBeg1:= Pointer(Name1);
    pBeg2:= Pointer(Name2);
    pEnd1 := pBeg1;
    pEnd2 := pBeg2;
    inc(pEnd1, Length(Name1));
    inc(pEnd2, Length(Name2));
    while (pEnd1 > pBeg1) and not CharInSet(pEnd1^, DELIMS) do dec(pEnd1);
    while (pEnd2 > pBeg2) and not CharInSet(pEnd2^, DELIMS) do dec(pEnd2);
    if pEnd1 > pBeg1 then inc(pEnd1);
    if pEnd2 > pBeg2 then inc(pEnd2);
    Result := (StrTools.StrIComp(pEnd1, pEnd2) = 0);
  end;
end;

function IsLegalFileName(const FileName: RawByteString): Longbool;
var
  iIndex: Integer;
begin
  if FileName <> E then begin
    Result := TRUE;
    for iIndex := 1 to Length(FileName) do begin
      if StrTools.CharInSet(FileName[iIndex], ILLEGAL_FN_CHARS) then begin
	Result := FALSE;
	Exit;
      end;
    end;
  end
  else Result := FALSE;
end;

function IsLegalFileName(const FileName: UnicodeString): Longbool;
var
  iIndex: Integer;
begin
  if FileName <> E then begin
    Result := TRUE;
    for iIndex := 1 to Length(FileName) do begin
      if CharInSet(FileName[iIndex], ILLEGAL_FN_CHARS) then begin
	Result := FALSE;
	Exit;
      end;
    end;
  end
  else Result := FALSE;
end;

function GetLegalFileName(const FileName: RawByteString): RawByteString;
var
  iLen, iIndex: Integer;
  pResult: PAnsiChar;
begin
  if FileName <> E then begin
    iLen := Length(FileName);
    SetString(Result, NIL, iLen);
    pResult := Pointer(Result);
    for iIndex := 1 to iLen do begin
      if StrTools.CharInSet(FileName[iIndex], ILLEGAL_FN_CHARS) then
	pResult^ := UD
      else pResult^ := FileName[iIndex];
      inc(pResult);
    end;
  end
  else Result := E;
end;

function GetLegalFileName(const FileName: UnicodeString): UnicodeString;
var
  iLen, iIndex: Integer;
  pResult: PWideChar;
begin
  if FileName <> E then begin
    iLen := Length(FileName);
    SetString(Result, NIL, iLen);
    pResult := Pointer(Result);
    for iIndex := 1 to iLen do begin
      if CharInSet(FileName[iIndex], ILLEGAL_FN_CHARS) then
	pResult^ := UD
      else pResult^ := FileName[iIndex];
      inc(pResult);
    end;
  end
  else Result := E;
end;

function GetLegalFileExt(const FileName: RawByteString): RawByteString;
begin
  Result := FileTools.ExtractFileExt(FileName);
  if not FileTools.IsLegalFileName(Result) then Result := E;
end;

function GetLegalFileExt(const FileName: UnicodeString): UnicodeString;
begin
  Result := FileTools.ExtractFileExt(FileName);
  if not IsLegalFileName(Result) then Result := E;
end;

procedure AddBackslash(var Path: RawByteString; PathLen: Integer);
var
  iLen: Integer;
begin
  if Path <> E then begin
    iLen := Length(Path);
    if (iLen < PathLen) or (PathLen <= 0) then PathLen := iLen;
    if Path[PathLen] <> BS then begin
      if PathLen < iLen then
	Path[succ(PathLen)] := BS
      else Path := Path + BS;
      Delete(Path, PathLen + 2, MAXINT);
    end
    else Delete(Path, succ(PathLen), MAXINT);
  end
  else Path := E;
end;

procedure AddBackslash(var Path: UnicodeString; PathLen: Integer);
var
  iLen: Integer;
begin
  if Path <> E then begin
    iLen := Length(Path);
    if (iLen < PathLen) or (PathLen <= 0) then PathLen := iLen;
    if Path[PathLen] <> BS then begin
      if PathLen < iLen then
	Path[succ(PathLen)] := BS
      else Path := Path + BS;
      Delete(Path, PathLen + 2, MAXINT);
    end
    else Delete(Path, succ(PathLen), MAXINT);
  end
  else Path := E;
end;

function AddBackslash(const Path: RawByteString): RawByteString;
begin
  Result := Path;
  AddBackslash(Result, Length(Result));
end;

function AddBackslash(const Path: UnicodeString): UnicodeString;
begin
  Result := Path;
  AddBackslash(Result, Length(Result));
end;

procedure RemoveBackslash(var Path: String; PathLen: Integer);
var
  iLen: Integer;
begin
  if Path <> E then begin
    iLen := Length(Path);
    if (iLen < PathLen) or (PathLen <= 0) then PathLen := iLen;
    while (PathLen > 0) and (Path[PathLen] = BS) do dec(PathLen);
    Delete(Path, succ(PathLen), MAXINT);
  end
  else Path := E;
end;

function RemoveBackslash(const Path: String): String; inline;
begin
  Result := Path;
  RemoveBackslash(Result, Length(Result));
end;

function GetAppNameFromCommandLine(const CommandLine: String): String;
begin
  Result := FileTools.ExtractFileName(GetAppPathFromCommandLine(CommandLine));
end;

function GetAppPathFromCommandLine(const CommandLine: String): String;
var
  iDelimiter: Integer;
begin
  iDelimiter := FirstStrPos(QU + SP, CommandLine);
  if iDelimiter = 0 then FirstCharPos(SP, CommandLine);
  if iDelimiter > 0 then begin
    Result := Unquote(Copy(CommandLine, 1, iDelimiter));
    Exit;
  end;
  Result := CommandLine;
end;

function ReplaceFileDrive(const Path, Drive: String): String;
begin
  Result := Drive + Copy(Path, succ(Length(FileTools.ExtractFileDrive(Path))), MAXINT);
end;

function ReplaceFileName(const Path, Name: String): String;
begin
  Result := AddBackSlash(FileTools.ExtractFilePath(Path)) + Name;
end;

function GetQuotedPath(const Path: String): String;
var
  LeadingQuote, TrailingQuote: String;
begin
  if Path <> E then begin
    LeadingQuote := E;
    if Path[1] <> QU then LeadingQuote := QU;
    TrailingQuote := E;
    if Path[Length(Path)] <> QU then TrailingQuote := QU;
    Result := LeadingQuote + Path + TrailingQuote;
  end
  else Result := Path;
end;

function GetShortPath(const Path: String): String;
var
  aPath: TUNCPath;
begin
  if Path <> E then begin
    SetString(Result, aPath, GetShortPathName(Pointer(Path), aPath, SizeOf(aPath)));
    Exit;
  end;
  Result := Path;
end;

function GetLongPath(const Path: String): String;
var
  aPath: TUNCPath;
begin
  if Path <> E then begin
    SetString(Result, aPath, GetLongPathName(Pointer(Path), aPath, SizeOf(aPath)));
    Exit;
  end;
  Result := Path;
end;

// returns trailing BS if one exists in Dir
function GetParentDir(const Dir: String): String;
var
  sEnd: String;
  aPath: TUNCPath;
  pFName: PWideChar;
begin
  if Dir <> E then begin
    if Dir[Length(Dir)] = BS then
      sEnd := DDOT + BS
    else sEnd := BS + DDOT;
    SetString(Result, aPath, GetFullPathName(PWideChar(Dir + sEnd), SizeOf(aPath), aPath, pFName));
    Exit;
  end;
  Result := E;
end;

function CreateSubDir(const Path: String): String;
var
  iPos: Integer;
  pResult: PChar;
begin
  iPos := 0;
  Result := Path;
  if (Result <> E) and not DirExists(Result) then begin
    UniqueString(Result);
    pResult := PChar(Result);
    repeat
      while not CharInSet(pResult[iPos], [NUL, BS]) do inc(iPos);
      if pResult[iPos] = NUL then Break;
      pResult[iPos] := NUL;
      try
	CreateDirectory(pResult, NIL);
      finally
	pResult[iPos] := BS;
      end;
      inc(iPos);
    until iPos = 0;
  end;
  if not DirExists(Result) then Result := E;
end;

function RemoveSubDirs(const FullPath, BasePath: String): String;
var
  sStop: String;
  iPos: Integer;
begin
  if (BasePath <> E) and StrTools.StartsText(BasePath, FullPath) then begin
    Result := AddBackslash(FullPath);
    sStop := AddBackslash(BasePath);
    repeat
      if RemoveDir(Result) then begin
	iPos := Length(Result);
	if (iPos > 0) and (Result[iPos] = BS) then dec(iPos);
	while (iPos > 0) and (Result[iPos] <> BS) do dec(iPos);
	Delete(Result, succ(iPos), MAXINT);
      end
      else begin
	Result := E;
	Exit;
      end;
    until StrTools.AnsiSameText(Result, sStop);
  end
  else Result := E;
end;

function SetCurDir(const Name: String): Longbool;
begin
  Result := SetCurrentDirectory(PChar(Name));
end;

function GetCurDir: String;
var
  iLen: Integer;
begin
  iLen := GetCurrentDirectory(0, NIL);
  if iLen > 0 then begin
    inc(iLen);
    SetString(Result, NIL, iLen);
    AddBackslash(Result, GetCurrentDirectory(iLen, PChar(Result)));
    Exit;
  end;
  Result := E;
end;

function GetWinDir: String;
begin
  Result := AddBackslash(GetSystemFolderPath(CSIDL_WINDOWS, FALSE));
end;

function GetSysDir: String;
begin
  Result := AddBackslash(GetSystemFolderPath(CSIDL_SYSTEMX86, FALSE));
end;

{$IFDEF UNICODE}
function GetTmpDir: UnicodeString; inline;
begin
  Result := GetTmpDirW;
end;
{$ELSE}
function GetTmpDir: RawByteString; inline;
begin
  Result := GetTmpDirA;
end;
{$ENDIF}

function GetTmpDirA: RawByteString;
var
  iLen: Integer;
begin
  iLen := GetTempPathA(0, NIL);
  if iLen > 0 then begin
    SetString(Result, NIL, pred(iLen));
    GetTempPathA(iLen, Pointer(Result));
    if not DirExists(Result) then CreateDirectoryA(PAnsiChar(Result), NIL);
    Exit;
  end;
  Result := E;
end;

function GetTmpDirW: UnicodeString;
var
  iLen: Integer;
begin
  iLen := GetTempPathW(0, NIL);
  if iLen > 0 then begin
    SetString(Result, NIL, pred(iLen));
    GetTempPathW(iLen, Pointer(Result));
    if not DirExists(Result) then CreateDirectoryW(PWideChar(Result), NIL);
    Exit;
  end;
  Result := E;
end;

function GetModuleName(ModuleHandle: THandle): String;
var
  aPath: TUNCPath;
begin
  SetString(Result, aPath, GetModuleFileName(ModuleHandle, aPath, SizeOf(aPath)));
end;

function GetModulePath(ModuleHandle: THandle): String;
begin
  Result := AddBackslash(FileTools.ExtractFilePath(GetModuleName(ModuleHandle)));
end;

function GetModuleOrAppPath(const FileName: String; const AppPath: String): String;
begin
  Result := GetModulePath(HInstance) + FileName;
  if FileExists(Result) then Exit;
  Result := GetModulePath(0) + FileName;
  if FileExists(Result) then Exit;
  Result := AppPath;
  if FileExists(Result) then Exit;
  Result := FileName;
  if FileExists(Result) then Exit;
  Result := AddBackSlash(AppPath) + FileName;
end;

function GetUniqueFileName(const FileExt: String; Hashed: Longbool = TRUE): String;
var
  stCurrent: TSystemTime;
  dtCurrent: TDateTime;
begin
  GetSystemTime(stCurrent);
  dtCurrent := SystemTimeToDateTime(stCurrent);
  if Hashed then begin
    DateTimeToString(Result, 'yyyymmddhhnnsszzz', dtCurrent);
    Result := HashedString(Result, 8) + FileExt;
  end
  else begin
    DateTimeToString(Result, 'yyyy-mm-dd-hh-nn-ss-zzz', dtCurrent);
    Result := Result + FileExt;
  end;
end;

function GetDriveType(const Name: String): TDriveType;
begin
  Result := TDriveType(Windows.GetDriveType(PChar(AddBackslash(FileTools.ExtractFileDrive(Name)))));
end;

function GetUNCFileName(const FileName: String): String;
var
  dwSize: DWord;
  bDummy: Byte;
  sName: String;
begin
  dwSize := 1;
  if WNetGetUniversalName(PChar(FileName), UNIVERSAL_NAME_INFO_LEVEL, @bDummy, dwSize) = ERROR_MORE_DATA then begin
    SetString(sName, NIL, dwSize);
    if WNetGetUniversalName(PChar(FileName), UNIVERSAL_NAME_INFO_LEVEL, Pointer(sName), dwSize) = 0 then begin
      Result := PUniversalNameInfo(sName).lpUniversalName;
      Exit;
    end;
  end;
  Result := FileName;
end;

function GetFileNameByHandle(FileHandle: THandle): String;
var
  pni: PNameInfo;
begin
  pni := AllocMem($FFFF);
  try
    if GetFileInformationByHandleEx(FileHandle, FileNameInfo, pni, $FFFF) then begin
      with pni^ do if FileNameLength > 0 then begin
	SetString(Result, NIL, FileNameLength div SizeOf(Char));
	Move(FileName[0], Pointer(Result)^, FileNameLength);
	Exit;
      end;
    end;
  finally
    FreeMem(pni);
  end;
  Result := E;
end;

function GetLocalizedPath(const Path: String): String;
var
  sTrimPath: String;
  iLen, iPos: Integer;
  fQuoted, fBS: Longbool;
  paPath: TPathArray;
  shfi: TSHFileInfo;
begin
  if not StrTools.StartsText(BS + BS, StrTools.Trim(Path)) then begin
    sTrimPath := StrTools.Unquote(Path, QU, fQuoted);
    iLen := Length(sTrimPath);
    if iLen = 0 then begin
      Result := E;
      Exit;
    end;
    fBS := (sTrimPath[iLen] = BS);
    StrTools.StrPLCopy(paPath, sTrimPath, MAX_PATH);
    if fBS then paPath[pred(iLen)] := NUL;
    if (iLen < 3) or (sTrimPath[2] <> CL) then begin
      Result := E;
      Exit;
    end;
    iPos := 3;
    Result := sTrimPath[1] + CL;
    repeat
      inc(iPos);
      while not CharInSet(sTrimPath[iPos], [NUL, BS]) do inc(iPos);
      paPath[pred(iPos)] := NUL;
      if SHGetFileInfo(paPath, DWord(-1), shfi, SizeOf(shfi), SHGFI_DISPLAYNAME) <> 0 then begin
	if Result <> E then
	  Result := Result + BS + shfi.szDisplayName
	else Result := Result + shfi.szDisplayName;
      end
      else begin
	Result := sTrimPath;
	Exit;
      end;
      paPath[pred(iPos)] := sTrimPath[iPos];
    until iPos >= iLen;
    if fBS then AddBackslash(Result, StrTools.StrLen(PChar(Result)));
    if fQuoted then Result := StrTools.AnsiQuotedStr(Result, QU);
  end
  else Result := sTrimPath;
end;

function GetLocaleDir(Base: Longbool = FALSE): String;
var
  iPos, iLen: Integer;
  lcid: TLCID;
begin
  if Base then begin
    Result := '\en-US';
    Exit;
  end;
  iPos := 1;
  lcid := LOCALE_USER_DEFAULT;
  iLen := GetLocaleInfo(lcid, LOCALE_SISO639LANGNAME, NIL, 0) +
	  GetLocaleInfo(lcid, LOCALE_SISO3166CTRYNAME, NIL, 0) + 1;
  SetString(Result, NIL, pred(iLen));
  Result[iPos] := BS;
  dec(iLen, iPos);
  inc(iPos, GetLocaleInfo(lcid, LOCALE_SISO639LANGNAME, @Result[succ(iPos)], succ(iLen)));
  Result[iPos] := DA;
  dec(iLen, iPos);
  GetLocaleInfo(lcid, LOCALE_SISO3166CTRYNAME, @Result[succ(iPos)], succ(iLen));
end;

function GetOrLoadLibrary(const Name: String): THandle;
begin
  Result := GetModuleHandle(PChar(Name));
  if Result = 0 then Result := LoadLibrary(PChar(Name));
end;

function LoadLibraryEx(const Name: String; Flags: Cardinal): THandle;
begin
  if (MainWinVersion < 6) and ((Flags and LOAD_LIBRARY_AS_DATAFILE_EXCLUSIVE) <> 0) then Flags := (Flags and $F) or LOAD_LIBRARY_AS_DATAFILE;
  Result := Windows.LoadLibraryEx(PChar(Name), 0, Flags);
end;

function LoadMuiLibrary(const LibInfo: Variant; Localize: Longbool): THandle;
var
  hLib: THandle;
  sSubDir, sLibDir: String;
begin
  case VarType(LibInfo) of
    varInteger: hLib := LibInfo;
    varUString: hLib := GetModuleHandle(PChar(String(LibInfo)));
  else
    hLib := 0;
  end;
  if hLib <> 0 then begin
    sLibDir := GetModuleName(hLib);
    sSubDir := GetLocaleDir(not Localize);
    Insert(sSubDir, sLibDir, succ(AnsiStrRScan(PChar(sLibDir), BS) - Pointer(sLibDir)));
    sLibDir := sLibDir + '.mui';
    Result := GetModuleHandle(PChar(sLibDir));
    if Result = 0 then begin
      Result := LoadLibraryEx(PChar(sLibDir), LOAD_LIBRARY_AS_DATAFILE);
      if Result = 0 then Result := hLib;
    end;
  end
  else Result := 0;
end;

function MakeErrorFileExtension(Code: Cardinal; Digits: THexDigits = 4): String;
begin
  Result := DOT + StrTools.Format('$%.' + Char(Ord('0') + Digits) + 'x', [Code]);
end;

function SetErrorFileExtension(const FileName: String; Code: Cardinal; Digits: THexDigits = 4): String;
begin
  Result := FileTools.ChangeFileExt(FileName, MakeErrorFileExtension(Code, Digits));
end;

function NoValidFile(const FileName: RawByteString): Longbool;
begin
  Result := (FileName = E) or ((FileName <> E) and (not FileExists(FileName) or (FileSize(FileName) = 0)));
end;

function NoValidFile(const FileName: UnicodeString): Longbool;
begin
  Result := (FileName = E) or ((FileName <> E) and (not FileExists(FileName) or (FileSize(FileName) = 0)));
end;

function ValidFile(const FileName: RawByteString): Longbool;
begin
  Result := not NoValidFile(FileName);
end;

function ValidFile(const FileName: UnicodeString): Longbool;
begin
  Result := not NoValidFile(FileName);
end;

function IsExecutable(const FileName: RawByteString): Longbool;
var
  SHFI: TSHFileInfoA;
begin
  Result := (LoWord(SHGetFileInfoA(PAnsiChar(FileName), 0, SHFI, SizeOf(SHFI), SHGFI_EXETYPE)) <> 0);
end;

function IsExecutable(const FileName: UnicodeString): Longbool;
var
  SHFI: TSHFileInfoW;
begin
  Result := (LoWord(SHGetFileInfoW(PWideChar(FileName), 0, SHFI, SizeOf(SHFI), SHGFI_EXETYPE)) <> 0);
end;

function IsBinaryBitness(const FileName: RawByteString): TBinaryBitness;
var
  dwBinType: DWord;
begin
  dwBinType := 0;
  if GetBinaryTypeA(PAnsiChar(UNC_PREFIX + FileName), dwBinType) then
    Result := TBinaryBitness(dwBinType)
  else Result := MAX_DWORD;
end;

function IsBinaryBitness(const FileName: UnicodeString): TBinaryBitness;
var
  dwBinType: DWord;
begin
  dwBinType := 0;
  if GetBinaryTypeW(PWideChar(UNC_PREFIX + FileName), dwBinType) then
    Result := TBinaryBitness(dwBinType)
  else Result := MAX_DWORD;
end;

function GetSystemFolderPath(FolderID: Integer; Localize: Longbool = FALSE): String;
var
  pidl: PItemIDList;
begin
  case FolderID of
    CSIDL_PROGRAM_FILES: Result := GetSystemFolderPath(FOLDERID_ProgramFilesX64);
    CSIDL_PROGRAM_FILES_COMMON: Result := GetSystemFolderPath(FOLDERID_ProgramFilesCommonX64);
  else
    Result := E;
  end;
  if (Result = E) and (SHGetSpecialFolderLocation(0, FolderID, pidl) = S_OK) then
    Result := GetSystemFolderPath(pidl, Localize)
  else if (FolderID = CSIDL_MYDOCUMENTS) and (SHGetSpecialFolderLocation(0, CSIDL_PERSONAL, pidl) = S_OK) then
    Result := GetSystemFolderPath(pidl, Localize)
  else Result := E;
end;

function GetSystemFolderPath(FolderID: PItemIDList; Localize: Longbool = FALSE): String;
begin
  SetString(Result, NIL, succ(MAX_PATH));
  if (FolderID <> NIL) and SHGetPathFromIDList(FolderID, PChar(Result)) then begin
    AddBackslash(Result, StrTools.StrLen(PChar(Result)));
    if Localize then Result := GetLocalizedPath(Result);
  end
  else Result := E;
end;

function GetSystemFolderPath(const KnownFolderID: TIID; Localize: Longbool = FALSE): String;
var
  wcFolder: PChar;
begin
  if (MainWinVersion >= 6) and (SHGetKnownFolderPath(KnownFolderID, 0, 0, wcFolder) = S_OK) then begin
    try
      if Localize then
	Result := AddBackslash(GetLocalizedPath(wcFolder))
      else Result := AddBackslash(wcFolder);
    finally
      CoTaskMemFree(wcFolder);
    end;
  end
  else Result := E;
end;

function CreateTempFileA(Path: RawByteString = E; Prefix: RawByteString = 'TMP'; Keep: Longbool = FALSE): RawByteString;
begin
  if Path = E then Path := GetTmpDirA;
  if Prefix = E then Prefix := 'TMP';
  SetString(Result, NIL, MAX_PATH);
  if GetTempFileNameA(PAnsiChar(Path), PAnsiChar(Prefix), 0, Pointer(Result)) <> 0 then begin
    if not Keep then Delete(Result, succ(StrTools.StrLen(PAnsiChar(Result))), MAXINT);
    Exit;
  end;
  Result := E;
end;

function CreateTempFileW(Path: UnicodeString = E; Prefix: UnicodeString = 'TMP'; Keep: Longbool = FALSE): UnicodeString;
begin
  if Path = E then Path := GetTmpDirW;
  if Prefix = E then Prefix := 'TMP';
  SetString(Result, NIL, MAX_PATH);
  if GetTempFileNameW(PWideChar(Path), PWideChar(Prefix), 0, Pointer(Result)) <> 0 then begin
    if not Keep then Delete(Result, succ(StrTools.StrLen(PWideChar(Result))), MAXINT);
    Exit;
  end;
  Result := E;
end;

function CreateTempFileByName(const FileName: RawByteString): RawByteString;
begin
  if CreateEmptyFile(FileName) <> E then begin
    Result := FileName;
    Exit;
  end;
  Result := E;
end;

function CreateTempFileByName(const FileName: UnicodeString): UnicodeString;
begin
  if CreateEmptyFile(FileName) <> E then begin
    Result := FileName;
    Exit;
  end;
  Result := E;
end;

function CreateLinkFile(const Link, Target, Commands, Icon: RawByteString): Longbool;
var
  iIcon, iPos: Integer;
  sIconFile: RawByteString;
  ShellLink: IShellLinkA;
  PersistFile: IPersistFile;
begin
  Result := FALSE;
  if (Link <> E) and (Target <> E) then begin
    iIcon := -1;
    if Icon <> E then begin
      iPos := LastCharPos(CM, Icon);
      if iPos > 0 then iIcon := StrTools.StrToInt64Def(StrTools.Copy(Icon, succ(iPos), MAXINT), -1);
      if iIcon > -1 then sIconFile := Copy(Icon, 1, pred(iPos));
    end;
    ShellLink := CreateComObject(CLSID_ShellLink) as IShellLinkA;
    if ShellLink <> NIL then with ShellLink do begin
      if SetPath(PAnsiChar(Target)) <> S_OK then Exit;
      if SetArguments(PAnsiChar(Commands)) <> S_OK then Exit;
      if (iIcon > -1) and (SetIconLocation(PAnsiChar(sIconFile), iIcon) <> S_OK) then Exit;
      if QueryInterface(IPersistFile, PersistFile) <> S_OK then Exit;
    end;
    Result := (PersistFile.Save(PWideChar(String(Link)), TRUE) = S_OK);
  end;
end;

function CreateLinkFile(const Link, Target, Commands, Icon: UnicodeString): Longbool;
var
  iIcon, iPos: Integer;
  sIconFile: UnicodeString;
  ShellLink: IShellLinkW;
  PersistFile: IPersistFile;
begin
  Result := FALSE;
  if (Link <> E) and (Target <> E) then begin
    iIcon := -1;
    if Icon <> E then begin
      iPos := LastCharPos(CM, Icon);
      if iPos > 0 then iIcon := StrTools.StrToInt64Def(Copy(Icon, succ(iPos), MAXINT), -1);
      if iIcon > -1 then sIconFile := Copy(Icon, 1, pred(iPos));
    end;
    ShellLink := CreateComObject(CLSID_ShellLink) as IShellLinkW;
    if ShellLink <> NIL then with ShellLink do begin
      if SetPath(PWideChar(Target)) <> S_OK then Exit;
      if SetArguments(PWideChar(Commands)) <> S_OK then Exit;
      if (iIcon > -1) and (SetIconLocation(PWideChar(sIconFile), iIcon) <> S_OK) then Exit;
      if QueryInterface(IPersistFile, PersistFile) <> S_OK then Exit;
    end;
    Result := (PersistFile.Save(PWideChar(String(Link)), TRUE) = S_OK);
  end;
end;

function CreateUrlShortcut(const FileName, Url: RawByteString): Longbool;
var
  UniformResourceLocator: IUniformResourceLocatorA;
  Persistfile: IPersistFile;
begin
  if CoCreateInstance(CLSID_InternetShortcut, NIL, CLSCTX_INPROC_SERVER, IID_IUniformResourceLocator, UniformResourceLocator) = S_OK then begin
    if (UniformResourceLocator <> NIL) and (UniformResourceLocator.SetURL(PAnsiChar(Url), 0) = S_OK) then begin
      Persistfile := UniformResourceLocator as IPersistFile;
      Result := (PersistFile.Save(PWideChar(String(FileName)), FALSE) = S_OK);
      Exit;
    end;
  end;
  Result := FALSE;
end;

function CreateUrlShortcut(const FileName, Url: UnicodeString): Longbool;
var
  UniformResourceLocator: IUniformResourceLocatorW;
  Persistfile: IPersistFile;
begin
  if CoCreateInstance(CLSID_InternetShortcut, NIL, CLSCTX_INPROC_SERVER, IID_IUniformResourceLocator, UniformResourceLocator) = S_OK then begin
    if (UniformResourceLocator <> NIL) and (UniformResourceLocator.SetURL(PWideChar(Url), 0) = S_OK) then begin
      Persistfile := UniformResourceLocator as IPersistFile;
      Result := (PersistFile.Save(PWideChar(String(FileName)), FALSE) = S_OK);
      Exit;
    end;
  end;
  Result := FALSE;
end;

function GetLinkTarget(const LinkFile: RawByteString): RawByteString; overload;
var
  ShellLink: IShellLinkA;
  PersistFile: IPersistFile;
  fdLinkFile: TWin32FindDataA;
begin
  if LinkFile <> E then begin
    ShellLink := CreateComObject(CLSID_ShellLink) as IShellLinkA;
    if (ShellLink <> NIL) and (ShellLink.QueryInterface(IPersistFile, PersistFile) = S_OK) then begin
      SetString(Result, NIL, MAX_UNC_PATH);
      if (PersistFile.Load(PWideChar(String(LinkFile)), STGM_READ) <> S_OK) or (ShellLink.Resolve(0, SLR_NO_UI) <> S_OK)
      or (ShellLink.GetPath(Pointer(Result), MAX_PATH, fdLinkFile, 0) = S_OK) then begin
	Delete(Result, succ(StrTools.StrLen(PAnsiChar(Result))), MAXINT);
	Exit;
      end;
    end;
  end;
  Result := E;
end;

function GetLinkTarget(const LinkFile: UnicodeString): UnicodeString; overload;
var
  ShellLink: IShellLinkW;
  PersistFile: IPersistFile;
  fdLinkFile: TWin32FindDataW;
begin
  if LinkFile <> E then begin
    ShellLink := CreateComObject(CLSID_ShellLink) as IShellLinkW;
    if (ShellLink <> NIL) and (ShellLink.QueryInterface(IPersistFile, PersistFile) = S_OK) then begin
      SetString(Result, NIL, MAX_UNC_PATH);
      if (PersistFile.Load(PWideChar(String(LinkFile)), STGM_READ) <> S_OK) or (ShellLink.Resolve(0, SLR_NO_UI) <> S_OK)
      or (ShellLink.GetPath(Pointer(Result), MAX_PATH, fdLinkFile, 0) = S_OK) then begin
	Delete(Result, succ(StrTools.StrLen(PWideChar(Result))), MAXINT);
	Exit;
      end;
    end;
  end;
  Result := E;
end;

function ReadWriteAccess(const FileOrFolder: RawByteString): Longbool;
begin
  Result := Longbool(CreateFileA(PAnsiChar(FileOrFolder), GENERIC_READ or GENERIC_WRITE, FILE_SHARE_READ or FILE_SHARE_WRITE,
		     NIL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, 0));
  if THandle(Result) <> INVALID_HANDLE_VALUE then
    CloseHandle(THandle(Result))
  else Result := FALSE;
end;

function ReadWriteAccess(const FileOrFolder: UnicodeString): Longbool;
begin
  Result := Longbool(CreateFileW(PWideChar(FileOrFolder), GENERIC_READ or GENERIC_WRITE, FILE_SHARE_READ or FILE_SHARE_WRITE,
		     NIL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, 0));
  if THandle(Result) <> INVALID_HANDLE_VALUE then
    CloseHandle(THandle(Result))
  else Result := FALSE;
end;

function ReadFromConsole: RawByteString;
var
  hConsole: THandle;
  dwSize, dwResult: DWord;
begin
  hConsole := GetStdHandle(STD_INPUT_HANDLE);
  if hConsole <> INVALID_HANDLE_VALUE then begin
    dwSize := SetFilePointer(hConsole, 0, NIL, FILE_END);
    SetString(Result, NIL, dwSize);
    SetFilePointer(hConsole, 0, NIL, FILE_BEGIN);
    if ReadFile(hConsole, Pointer(Result)^, dwSize, dwResult, NIL) and (dwResult = dwSize) then Exit;
  end;
  Result := E;
end;

procedure WriteToConsole(const Text: RawByteString);
var
  hConsole: THandle;
  dwSize, dwResult: DWord;
begin
  hConsole := GetStdHandle(STD_OUTPUT_HANDLE);
  if hConsole <> INVALID_HANDLE_VALUE then begin
    dwResult := 0;
    dwSize := Length(Text) * SizeOf(AnsiChar);
    if WriteFile(hConsole, PAnsiChar(Text)^, dwSize, dwResult, NIL) and (dwResult = dwSize) then SetEndOfFile(hConsole);
  end;
end;

procedure WriteToConsole(const Text: UnicodeString);
var
  hConsole: THandle;
  dwSize, dwResult: DWord;
begin
  hConsole := GetStdHandle(STD_OUTPUT_HANDLE);
  if hConsole <> INVALID_HANDLE_VALUE then begin
    dwResult := 0;
    dwSize := Length(Text) * SizeOf(WideChar);
    if WriteFile(hConsole, PWideChar(Text)^, dwSize, dwResult, NIL) and (dwResult = dwSize) then SetEndOfFile(hConsole);
  end;
end;

function CopyFile(const Source, Dest: RawByteString; FailIfExists: Longbool): Longbool;
begin
  Result := StrTools.SameText(Source, Dest) or CopyFileA(PAnsiChar(Source), PAnsiChar(Dest), FailIfExists);
end;

function CopyFile(const Source, Dest: UnicodeString; FailIfExists: Longbool): Longbool;
begin
  Result := StrTools.SameText(Source, Dest) or CopyFileW(PWideChar(Source), PWideChar(Dest), FailIfExists);
end;

function MoveFile(const Source, Dest: RawByteString; Overwrite: Longbool): Longbool;
begin
  if Overwrite and not StrTools.SameText(Source, Dest) then SetFileReadOnly(Dest, FALSE);
  Result := StrTools.SameText(Source, Dest) or
    MoveFileExA(PAnsiChar(Source), PAnsiChar(Dest), MOVEFILE_REPLACE_EXISTING * Ord(Overwrite = TRUE) or
						    MOVEFILE_COPY_ALLOWED or MOVEFILE_WRITE_THROUGH);
end;

function MoveFile(const Source, Dest: UnicodeString; Overwrite: Longbool): Longbool;
begin
  if Overwrite and not StrTools.SameText(Source, Dest) then SetFileReadOnly(Dest, FALSE);
  Result := StrTools.SameText(Source, Dest) or
	    MoveFileExW(PWideChar(Source), PWideChar(Dest), MOVEFILE_REPLACE_EXISTING * Ord(Overwrite = TRUE) or
							    MOVEFILE_COPY_ALLOWED or MOVEFILE_WRITE_THROUGH);
end;

function IsFileReadOnly(const FileName: RawByteString): Longbool;
var
  dwCode: DWord;
begin
  dwCode := GetFileAttributesA(Pointer(FileName));
  Result := (dwCode <> INVALID_FILE_ATTRIBUTES) and (dwCode and FILE_ATTRIBUTE_READONLY <> 0);
end;

function IsFileReadOnly(const FileName: UnicodeString): Longbool;
var
  dwCode: DWord;
begin
  dwCode := GetFileAttributesW(Pointer(FileName));
  Result := (dwCode <> INVALID_FILE_ATTRIBUTES) and (dwCode and FILE_ATTRIBUTE_READONLY <> 0);
end;

function SetFileReadOnly(const FileName: RawByteString; Value: Boolean): Boolean;
var
  dwCode: DWord;
begin
  dwCode := GetFileAttributesA(Pointer(FileName));
  if dwCode <> INVALID_FILE_ATTRIBUTES then begin
    Result := (dwCode and FILE_ATTRIBUTE_READONLY <> 0);
    if Result <> Value then begin
      case Value of
	TRUE: dwCode := dwCode or FILE_ATTRIBUTE_READONLY;
	FALSE: dwCode := dwCode and not FILE_ATTRIBUTE_READONLY;
      end;
      Result := SetFileAttributesA(Pointer(FileName), dwCode);
    end
    else Result := TRUE;
    Exit;
  end;
  Result := FALSE;
end;

function SetFileReadOnly(const FileName: UnicodeString; Value: Boolean): Boolean;
var
  dwCode: DWord;
begin
  dwCode := GetFileAttributesW(Pointer(FileName));
  if dwCode <> INVALID_FILE_ATTRIBUTES then begin
    Result := (dwCode and FILE_ATTRIBUTE_READONLY <> 0);
    if Result <> Value then begin
      case Value of
	TRUE: dwCode := dwCode or FILE_ATTRIBUTE_READONLY;
	FALSE: dwCode := dwCode and not FILE_ATTRIBUTE_READONLY;
      end;
      Result := SetFileAttributesW(Pointer(FileName), dwCode);
    end
    else Result := TRUE;
    Exit;
  end;
  Result := FALSE;
end;

function RenameWriteProtect(const Old, New: RawByteString; Overwrite, Protect: Boolean): Longbool;
begin
  Result := SetFileReadOnly(Old, FALSE) and MoveFile(Old, New, Overwrite);
  if Result and Protect then Result := SetFileReadOnly(New, TRUE);
end;

function RenameWriteProtect(const Old, New: UnicodeString; Overwrite, Protect: Boolean): Longbool;
begin
  Result := SetFileReadOnly(Old, FALSE) and MoveFile(Old, New, Overwrite);
  if Result and Protect then Result := SetFileReadOnly(New, TRUE);
end;

function AppendTextFile(const Name: RawByteString; const Text: RawByteString; AsLine: Longbool = false): Longbool;
var
  hFile: THandle;
  dwSize, dwResult: DWord;
begin
  Result := FALSE;
  hFile := CreateOrOpenFile(Name);
  if hFile <> INVALID_HANDLE_VALUE then begin
    try
      dwSize := Length(Text);
      if AsLine then inc(dwSize, Length(CRLF));
      dwSize := dwSize * SizeOf(AnsiChar);
      SetFilePointer(hFile, 0, NIL, FILE_END);
      Result := WriteFile(hFile, PAnsiChar(Text + CRLF)^, dwSize, dwResult, NIL) and (dwResult = dwSize);
    finally
      SetEndOfFile(hFile);
      CloseHandle(hFile);
    end;
  end;
end;

function AppendTextFile(const Name: String; const Text: RawByteString; AsLine: Longbool = false): Longbool;
var
  hFile: THandle;
  dwSize, dwResult: DWord;
begin
  Result := FALSE;
  hFile := CreateOrOpenFile(Name);
  if hFile <> INVALID_HANDLE_VALUE then begin
    try
      dwSize := Length(Text);
      if AsLine then inc(dwSize, Length(CRLF));
      dwSize := dwSize * SizeOf(AnsiChar);
      SetFilePointer(hFile, 0, NIL, FILE_END);
      Result := WriteFile(hFile, PAnsiChar(Text + CRLF)^, dwSize, dwResult, NIL) and (dwResult = dwSize);
    finally
      SetEndOfFile(hFile);
      CloseHandle(hFile);
    end;
  end;
end;

function AppendTextFile(const Name: String; const Text: UnicodeString; AsLine: Longbool = false): Longbool;
var
  hFile: THandle;
  dwSize, dwResult: DWord;
begin
  Result := FALSE;
  hFile := CreateOrOpenFile(Name);
  if hFile <> INVALID_HANDLE_VALUE then begin
    try
      dwSize := Length(Text);
      if AsLine then inc(dwSize, Length(CRLF));
      dwSize := dwSize * SizeOf(WideChar);
      SetFilePointer(hFile, 0, NIL, FILE_END);
      Result := WriteFile(hFile, PWideChar(Text + CRLF)^, dwSize, dwResult, NIL) and (dwResult = dwSize);
    finally
      SetEndOfFile(hFile);
      CloseHandle(hFile);
    end;
  end;
end;

function WriteTextFile(const Name: RawByteString; const Text: RawByteString): Longbool;
var
  hFile: THandle;
  dwSize, dwResult: DWord;
begin
  Result := FALSE;
  hFile := CreateWriteFile(Name);
  if hFile <> INVALID_HANDLE_VALUE then begin
    try
      dwResult := 0;
      dwSize := Length(Text);
      Result := WriteFile(hFile, PAnsiChar(Text)^, dwSize, dwResult, NIL) and (dwResult = dwSize);
    finally
      SetEndOfFile(hFile);
      CloseHandle(hFile);
    end;
  end;
end;

function WriteTextFile(const Name: String; const Text: RawByteString): Longbool;
var
  hFile: THandle;
  dwSize, dwResult: DWord;
begin
  Result := FALSE;
  hFile := CreateWriteFile(Name);
  if hFile <> INVALID_HANDLE_VALUE then begin
    try
      dwResult := 0;
      dwSize := Length(Text);
      Result := WriteFile(hFile, PAnsiChar(Text)^, dwSize, dwResult, NIL) and (dwResult = dwSize);
    finally
      SetEndOfFile(hFile);
      CloseHandle(hFile);
    end;
  end;
end;

function WriteTextFile(const Name: String; const Text: UnicodeString): Longbool;
var
  hFile: THandle;
  dwSize, dwResult: DWord;
begin
  Result := FALSE;
  hFile := CreateWriteFile(Name);
  if hFile <> INVALID_HANDLE_VALUE then begin
    try
      dwResult := 0;
      dwSize := Length(Text) * SizeOf(WideChar);
      Result := WriteFile(hFile, PWideChar(Text)^, dwSize, dwResult, NIL) and (dwResult = dwSize);
    finally
      SetEndOfFile(hFile);
      CloseHandle(hFile);
    end;
  end;
end;

function WriteTempFile(const Name, Prefix: RawByteString; const Text: RawByteString): RawByteString;
var
  hFile: THandle;
  dwSize, dwResult: DWord;
begin
  Result := CreateTempFileA(Name, Prefix);
  if Result <> E then begin
    hFile := CreateWriteFile(Result);
    try
      dwSize := Length(Text);
      if not WriteFile(hFile, Pointer(Text)^, dwSize, dwResult, NIL) or (dwResult <> dwSize) then begin
        CloseHandle(hFile);
        DeleteFile(Result);
	Result := E;
	Exit;
      end;
    finally
      CloseHandle(hFile);
    end;
  end;
end;

function WriteTempFile(const Name, Prefix: String; const Text: RawByteString): UnicodeString;
var
  hFile: THandle;
  dwSize, dwResult: DWord;
begin
  Result := CreateTempFileW(Name, Prefix);
  if Result <> E then begin
    hFile := CreateWriteFile(Result);
    try
      dwSize := Length(Text);
      if not WriteFile(hFile, Pointer(Text)^, dwSize, dwResult, NIL) or (dwResult <> dwSize) then begin
        CloseHandle(hFile);
        DeleteFile(Result);
	Result := E;
	Exit;
      end;
    finally
      CloseHandle(hFile);
    end;
  end;
end;

function WriteTempFile(const Name, Prefix: String; const Text: UnicodeString): UnicodeString;
var
  hFile: THandle;
  dwSize, dwResult: DWord;
begin
  Result := CreateTempFileW(Name, Prefix);
  if Result <> E then begin
    hFile := CreateWriteFile(Result);
    try
      dwSize := Length(Text) * SizeOf(WideChar);
      if not WriteFile(hFile, Pointer(Text)^, dwSize, dwResult, NIL) or (dwResult <> dwSize) then begin
        CloseHandle(hFile);
        DeleteFile(Result);
	Result := E;
	Exit;
      end;
    finally
      CloseHandle(hFile);
    end;
  end;
end;

function ShowTextFile(const Name: RawByteString; const Text: RawByteString): Longbool;
begin
  Result := WriteTextFile(Name, Text);
  if Result then Result := (ShellExecuteA(0, 'open', Pointer(Name), NIL, NIL, SW_SHOWNORMAL) > 32);
end;

function ShowTextFile(const Name: String; const Text: RawByteString): Longbool;
begin
  Result := WriteTextFile(Name, Text);
  if Result then Result := (ShellExecuteW(0, 'open', Pointer(Name), NIL, NIL, SW_SHOWNORMAL) > 32);
end;

function ShowTextFile(const Name: String; const Text: UnicodeString): Longbool;
begin
  Result := WriteTextFile(Name, Text);
  if Result then Result := (ShellExecute(0, 'open', PChar(Name), NIL, NIL, SW_SHOWNORMAL) > 32);
end;

function ReadTextFile(const Name: RawByteString; Bytes: Integer = 0; CheckBom: Longbool = FALSE): RawByteString; overload;
var
  hFile: THandle;
  rsBom: RawByteString;
  dwSkip, dwSize, dwResult: DWord;
begin
  hFile := OpenReadFile(Name);
  if hFile <> INVALID_HANDLE_VALUE then begin
    try
      dwSkip := 0;
      dwSize := SetFilePointer(hFile, 0, NIL, FILE_END);
      if (Bytes <> 0) and (DWord(Abs(Bytes)) < dwSize) then dwSize := Abs(Bytes);
      if CheckBom then begin
	SetFilePointer(hFile, 0, NIL, FILE_BEGIN);
	rsBom := AnsiString(StringOfChar(NUL, Length(UTF_8_BOM)));
	if ReadFile(hFile, PAnsiChar(rsBom)^, Length(rsBom), dwResult, NIL) and (dwResult = DWord(Length(rsBom))) then begin
	  case rsBom[1] of
	    #$EF: if StrAtPos(PAnsiChar(UTF_8_BOM), Pointer(rsBom), 1) then dwSkip := Length(UTF_8_BOM);
	    #$FE: if StrAtPos(PAnsiChar(UTF_BE_BOM), Pointer(rsBom), 1) then dwSkip := Length(UTF_BE_BOM);
	    #$FF: if StrAtPos(PAnsiChar(UTF_LE_BOM), Pointer(rsBom), 1) then dwSkip := Length(UTF_LE_BOM);
	  end;
	end;
      end;
      dec(dwSize, dwSkip);
      SetString(Result, NIL, dwSize);
      if Bytes >= 0 then
	SetFilePointer(hFile, dwSkip, NIL, FILE_BEGIN)
      else SetFilePointer(hFile, Bytes, NIL, FILE_END);
      if ReadFile(hFile, Pointer(Result)^, dwSize, dwResult, NIL) and (dwResult = dwSize) then Exit;
    finally
      CloseHandle(hFile);
    end;
  end;
  Result := E;
end;

function ReadTextFile(const Name: UnicodeString; Bytes: Integer = 0; CheckBom: Longbool = FALSE): RawByteString;
var
  hFile: THandle;
  rsBom: RawByteString;
  dwSkip, dwSize, dwResult: DWord;
begin
  hFile := OpenReadFile(Name);
  if hFile <> INVALID_HANDLE_VALUE then begin
    try
      dwSkip := 0;
      dwSize := SetFilePointer(hFile, 0, NIL, FILE_END);
      if (Bytes <> 0) and (DWord(Abs(Bytes)) < dwSize) then dwSize := Abs(Bytes);
      if CheckBom then begin
	SetFilePointer(hFile, 0, NIL, FILE_BEGIN);
	rsBom := AnsiString(StringOfChar(NUL, Length(UTF_8_BOM)));
	if ReadFile(hFile, PAnsiChar(rsBom)^, Length(rsBom), dwResult, NIL) and (dwResult = DWord(Length(rsBom))) then begin
	  case rsBom[1] of
	    #$EF: if StrAtPos(PAnsiChar(UTF_8_BOM), Pointer(rsBom), 1) then dwSkip := Length(UTF_8_BOM);
	    #$FE: if StrAtPos(PAnsiChar(UTF_BE_BOM), Pointer(rsBom), 1) then dwSkip := Length(UTF_BE_BOM);
	    #$FF: if StrAtPos(PAnsiChar(UTF_LE_BOM), Pointer(rsBom), 1) then dwSkip := Length(UTF_LE_BOM);
	  end;
	end;
      end;
      dec(dwSize, dwSkip);
      SetString(Result, NIL, dwSize);
      if Bytes >= 0 then
	SetFilePointer(hFile, dwSkip, NIL, FILE_BEGIN)
      else SetFilePointer(hFile, Bytes, NIL, FILE_END);
      if ReadFile(hFile, Pointer(Result)^, dwSize, dwResult, NIL) and (dwResult = dwSize) then Exit;
    finally
      CloseHandle(hFile);
    end;
  end;
  Result := E;
end;

{$WARNINGS OFF}
function ReadTextFileAsUtf8(const Name: String; const Charset: RawByteString; IncludeBom: Longbool = FALSE): Utf8String; inline;
begin
  Result := PAnsiChar(ReadTextFileAsUnicode(Name, Charset, TRUE, IncludeBom));
end;
{$WARNINGS ON}

function ReadTextFileAsUnicode(const Name: String; Charset: RawByteString; Utf8: Longbool = FALSE; IncludeBom: Longbool = FALSE): String;
var
  hFile: THandle;
  dwSize, dwResult: DWord;
  rsBom, rsFileData: RawByteString;
  fChanged: Boolean;
begin
  hFile := OpenReadFile(Name);
  if hFile <> INVALID_HANDLE_VALUE then begin
    try
      dwSize := SetFilePointer(hFile, 0, NIL, FILE_END);
      if dwSize > 0 then begin
	SetFilePointer(hFile, 0, NIL, FILE_BEGIN);
	rsBom := AnsiString(StringOfChar(NUL, Length(UTF_8_BOM)));
	if ReadFile(hFile, PAnsiChar(rsBom)^, Length(rsBom), dwResult, NIL) and (dwResult = DWord(Length(rsBom))) then begin
	  case rsBom[1] of
	    #$EF: if StrAtPos(PAnsiChar(UTF_8_BOM), Pointer(rsBom), 1) then Charset := UTF_8;
	    #$FE: if StrAtPos(PAnsiChar(UTF_BE_BOM), Pointer(rsBom), 1) then Charset := UTF_BE;
	    #$FF: if StrAtPos(PAnsiChar(UTF_LE_BOM), Pointer(rsBom), 1) then Charset := UTF_LE;
	  else
	    rsBom := E;
	  end;
	end;
	SetString(rsFileData, NIL, dwSize);
	SetFilePointer(hFile, 0, NIL, FILE_BEGIN);
	if ReadFile(hFile, Pointer(rsFileData)^, dwSize, dwResult, NIL) and (dwResult = dwSize) then begin
	  if Charset = UTF_16 then begin
	    if (rsBom = UTF_BE_BOM) or (rsFileData[1] = NUL) then
	      Charset := UTF_BE
	    else if (rsBom = UTF_LE_BOM) or (rsFileData[2] = NUL) then
	      Charset := UTF_LE;
	  end;
	  Result := FixUnicodeLineBreaks(AnyStringToWide(rsFileData, Charset, IncludeBom), fChanged);
	  if Utf8 then begin
	    if Result = E then
	      rsFileData := AnyStringToUtf8(rsFileData, Charset, IncludeBom)
	    else if fChanged or (Charset <> UTF_8) then rsFileData := WideStringToUtf8(Result);
	    if rsFileData <> E then begin
	      SetString(Result, NIL, DivModPlus(Length(rsFileData), 2));
	      Move(Pointer(rsFileData)^, Pointer(Result)^, succ(Length(rsFileData)));
	    end;
	  end;
	  Exit;
	end;
      end;
    finally
      CloseHandle(hFile);
    end;
  end;
  Result := E;
end;

function CreateOrOpenFile(const Name: RawByteString): THandle;
begin
  Result := CreateFileA(PAnsiChar(Name), GENERIC_READ or GENERIC_WRITE, FILE_SHARE_READ, NIL, OPEN_ALWAYS, FILE_FLAG_WRITE_THROUGH, 0);
end;

function CreateOrOpenFile(const Name: String): THandle;
begin
  Result := CreateFileW(PWideChar(Name), GENERIC_READ or GENERIC_WRITE, FILE_SHARE_READ, NIL, OPEN_ALWAYS, FILE_FLAG_WRITE_THROUGH, 0);
end;

function CreateWriteFile(const Name: RawByteString): THandle;
begin
  Result := CreateFileA(PAnsiChar(Name), GENERIC_READ or GENERIC_WRITE, FILE_SHARE_READ, NIL, CREATE_ALWAYS, FILE_FLAG_WRITE_THROUGH, 0);
end;

function CreateWriteFile(const Name: String): THandle;
begin
  Result := CreateFileW(PWideChar(Name), GENERIC_READ or GENERIC_WRITE, FILE_SHARE_READ, NIL, CREATE_ALWAYS, FILE_FLAG_WRITE_THROUGH, 0);
end;

function OpenWriteFile(const Name: RawByteString): THandle; overload;
begin
  Result := CreateFileA(PAnsiChar(Name), GENERIC_READ or GENERIC_WRITE, FILE_SHARE_READ, NIL, OPEN_EXISTING, FILE_FLAG_WRITE_THROUGH, 0);
end;

function OpenWriteFile(const Name: String): THandle; overload;
begin
  Result := CreateFileW(PWideChar(Name), GENERIC_READ or GENERIC_WRITE, FILE_SHARE_READ, NIL, OPEN_EXISTING, FILE_FLAG_WRITE_THROUGH, 0);
end;

function CreateEmptyFile(const Name: RawByteString): RawByteString;
var
  hFile: THandle;
begin
  hFile := CreateWriteFile(Name);
  if hFile <> INVALID_HANDLE_VALUE then begin
    CloseHandle(hFile);
    Result := Name;
    Exit;
  end;
  Result := E;
end;

function CreateEmptyFile(const Name: String): UnicodeString;
var
  hFile: THandle;
begin
  hFile := CreateWriteFile(Name);
  if hFile <> INVALID_HANDLE_VALUE then begin
    CloseHandle(hFile);
    Result := Name;
    Exit;
  end;
  Result := E;
end;

function TestWriteFile(const Name: RawByteString): Longbool;
var
  hFile: THandle;
begin
  hFile := OpenWriteFile(Name);
  if hFile <> INVALID_HANDLE_VALUE then begin
    CloseHandle(hFile);
    Result := TRUE;
  end
  else Result := FALSE;
end;

function TestWriteFile(const Name: String): Longbool;
var
  hFile: THandle;
begin
  hFile := OpenWriteFile(Name);
  if hFile <> INVALID_HANDLE_VALUE then begin
    CloseHandle(hFile);
    Result := TRUE;
  end
  else Result := FALSE;
end;

function OpenReadFile(const Name: RawByteString): THandle;
begin
  Result := CreateFileA(PAnsiChar(Name), GENERIC_READ, FILE_SHARE_READ, NIL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, 0);
end;

function OpenReadFile(const Name: UnicodeString): THandle;
begin
  Result := CreateFileW(PWideChar(Name), GENERIC_READ, FILE_SHARE_READ, NIL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, 0);
end;

function CloseFile(const Handle: THandle): Longbool;
begin
  Result := CloseHandle(Handle);
end;

function FileAge(const Name: RawByteString): TDateTime;
begin
  {$WARN SYMBOL_DEPRECATED OFF}
  if not SysUtils.FileAge(String(Name), Result) then Result := 0;
  {$WARN SYMBOL_DEPRECATED ON}
end;

function FileAge(const Name: UnicodeString): TDateTime;
begin
  {$WARN SYMBOL_DEPRECATED OFF}
  if not SysUtils.FileAge(String(Name), Result) then Result := 0;
  {$WARN SYMBOL_DEPRECATED ON}
end;

function FileSize(const Name: RawByteString): Int64;
var
  hFile: THandle;
begin
  hFile := OpenReadFile(Name);
  if hFile <> INVALID_HANDLE_VALUE then begin
    try
      if GetFileSizeEx(hFile, LARGE_INTEGER(Result)) then Exit;
    finally
      CloseHandle(hFile);
    end;
  end;
  Result := 0;
end;

function FileSize(const Name: UnicodeString): Int64;
var
  hFile: THandle;
begin
  hFile := OpenReadFile(Name);
  if hFile <> INVALID_HANDLE_VALUE then begin
    try
      if GetFileSizeEx(hFile, LARGE_INTEGER(Result)) then Exit;
    finally
      CloseHandle(hFile);
    end;
  end;
  Result := 0;
end;

function GetFileTypeInfo(const FileExtension: RawByteString; fti: TFileTypeInfos): RawByteString;
var
  dwLen: DWord;
begin
  dwLen := 0;
  Result := E;
  if AssocQueryStringA(0, Ord(fti), PAnsiChar(FileExtension), NIL, NIL, @dwLen) = S_FALSE then begin
    SetString(Result, NIL, pred(dwLen));
    if AssocQueryStringA(0, Ord(fti), PAnsiChar(FileExtension), NIL, Pointer(Result), @dwLen) = S_OK then Exit;
  end;
end;

function GetFileTypeInfo(const FileExtension: UnicodeString; fti: TFileTypeInfos): UnicodeString;
var
  dwLen: DWord;
begin
  dwLen := 0;
  Result := E;
  if AssocQueryStringW(0, Ord(fti), PWideChar(FileExtension), NIL, NIL, @dwLen) = S_FALSE then begin
    SetString(Result, NIL, pred(dwLen));
    if AssocQueryStringW(0, Ord(fti), PWideChar(FileExtension), NIL, Pointer(Result), @dwLen) = S_OK then Exit;
  end;
end;

function UpdateFileAccessTime(const Name: RawByteString; Time: TDateTime = 0): Longbool;
var
  hFile: THandle;
  stLocal: TSystemTime;
  ftLocal, ftSystem: TFileTime;
begin
  hFile := CreateFileA(PAnsiChar(Name), FILE_WRITE_ATTRIBUTES, FILE_SHARE_READ, NIL, OPEN_EXISTING, 0, 0);
  if hFile <> INVALID_HANDLE_VALUE then begin
    try
      if Time = 0 then Time := Now;
      DateTimeToSystemTime(Time, stLocal);
      Result := SystemTimeToFileTime(stLocal, ftLocal) and
		LocalFileTimeToFileTime(ftLocal, ftSystem) and
		SetFileTime(hFile, NIL, @ftSystem, NIL);
    finally
      CloseHandle(hFile);
    end;
  end
  else Result := FALSE;
end;

function UpdateFileAccessTime(const Name: UnicodeString; Time: TDateTime = 0): Longbool;
var
  hFile: THandle;
  stLocal: TSystemTime;
  ftLocal, ftSystem: TFileTime;
begin
  hFile := CreateFileW(PWideChar(Name), FILE_WRITE_ATTRIBUTES, FILE_SHARE_READ, NIL, OPEN_EXISTING, 0, 0);
  if hFile <> INVALID_HANDLE_VALUE then begin
    try
      if Time = 0 then Time := Now;
      DateTimeToSystemTime(Time, stLocal);
      Result := SystemTimeToFileTime(stLocal, ftLocal) and
		LocalFileTimeToFileTime(ftLocal, ftSystem) and
		SetFileTime(hFile, NIL, @ftSystem, NIL);
    finally
      CloseHandle(hFile);
    end;
  end
  else Result := FALSE;
end;

function FileVersionInfo(var FileVersionInfo: TFileVersionInfo): Longbool;
const
  COMMENTS_		= '\Comments';
  COMPANY_NAME		= '\CompanyName';
  FILE_DESCRIPTION	= '\FileDescription';
  FILE_VERSION		= '\FileVersion';
  INTERNAL_NAME		= '\InternalName';
  LEGAL_COPYRIGHT	= '\LegalCopyright';
  LEGAL_TRADEMARKS	= '\LegalTrademarks';
  ORIGINAL_FILENAME	= '\OriginalFilename';
  PRIVATE_BUILD		= '\PrivateBuild';
  PRODUCT_NAME		= '\ProductName';
  PRODUCT_VERSION	= '\ProductVersion';
  SPECIAL_BUILD		= '\SpecialBuild';

  function VerValue(const Info: RawByteString; const Name: String; var Data: Pointer): Cardinal;
  begin
    Result := 0;
    if not VerQueryValue(Pointer(Info), PChar(Name), Data, Result) then Exit;
  end;

  function VerFixed(const Info: RawByteString): TVSFixedFileInfo;
  var
    pffi: PVSFixedFileInfo;
  begin
    if VerValue(Info, BS, Pointer(pffi)) > 0 then begin
      Result := pffi^;
      Exit;
    end;
    FillChar(Result, SizeOf(TVSFixedFileInfo), 0);
  end;

  function VerTranslation(const Info: RawByteString): String;
  const
    VFI_TRANSLATION = '\VarFileInfo\Translation';
  type
    PTranslationInfo = ^TTranslationInfo;
    TTranslationInfo = Record
      wLanguage: Word;
      wCodePage: Word;
    end;
  var
    pti: PTranslationInfo;
  begin
    if VerValue(Info, VFI_TRANSLATION, Pointer(pti)) > 0 then begin
      with pti^ do Result := IntToHex(wLanguage shl 16 + wCodePage, 8);
      Exit;
    end;
    Result := E;
  end;

  function VerString(const Info: RawByteString; const Language, Value: String): String;
  const
    FILE_INFO = '\StringFileInfo\';
  var
    p: Pointer;
    c: Cardinal;
  begin
    c := VerValue(Info, FILE_INFO + Language + Value, p);
    if c > 0 then begin
      SetString(Result, PChar(p), pred(c));
      Exit;
    end;
    Result := E;
  end;

  function VersionInfo(const FileName: String): RawByteString;
  var
    dwSize, dwInfo: DWord;
  begin
    dwSize := GetFileVersionInfoSize(PChar(FileName), dwInfo);
    if dwSize > 0 then begin
      SetString(Result, NIL, dwSize);
      if GetFileVersionInfo(PChar(FileName), dwInfo, dwSize, Pointer(Result)) then Exit;
    end;
    Result := E;
  end;

var
  rsInfo: RawByteString;
begin
  Result := FALSE;
  with FileVersionInfo do if QueryFile <> E then begin
    rsInfo := VersionInfo(QueryFile);
    Result := (rsInfo <> E);
    if not Result then Exit;
    if fviFixed in QueryParameters then Fixed := VerFixed(rsInfo);
    if QueryLanguage = E then QueryLanguage := VerTranslation(rsInfo);
    if fviComments in QueryParameters then Comments := VerString(rsInfo, QueryLanguage, COMMENTS_);
    if fviCompanyName in QueryParameters then CompanyName := VerString(rsInfo, QueryLanguage, COMPANY_NAME);
    if fviFileDescription in QueryParameters then FileDescription := VerString(rsInfo, QueryLanguage, FILE_DESCRIPTION);
    if fviFileVersion in QueryParameters then FileVersion := VerString(rsInfo, QueryLanguage, FILE_VERSION);
    if fviInternalName in QueryParameters then InternalName := VerString(rsInfo, QueryLanguage, INTERNAL_NAME);
    if fviLegalCopyright in QueryParameters then LegalCopyright := VerString(rsInfo, QueryLanguage, LEGAL_COPYRIGHT);
    if fviLegalTrademarks in QueryParameters then LegalTrademarks := VerString(rsInfo, QueryLanguage, LEGAL_TRADEMARKS);
    if fviOriginalFilename in QueryParameters then OriginalFilename := VerString(rsInfo, QueryLanguage, ORIGINAL_FILENAME);
    if fviPrivateBuild in QueryParameters then PrivateBuild := VerString(rsInfo, QueryLanguage, PRIVATE_BUILD);
    if fviProductName in QueryParameters then ProductName := VerString(rsInfo, QueryLanguage, PRODUCT_NAME);
    if fviProductVersion in QueryParameters then ProductVersion := VerString(rsInfo, QueryLanguage, PRODUCT_VERSION);
    if fviSpecialBuild in QueryParameters then SpecialBuild := VerString(rsInfo, QueryLanguage, SPECIAL_BUILD);
  end;
end;

function TextAndMediaOnly(const Name: String): Longbool;
begin
  Result := StrTools.StartsText('audio', Name) or StrTools.StartsText('image', Name)
  or StrTools.StartsText('midi', Name) or StrTools.StartsText('video', Name) or StrTools.StartsText('text', Name);
end;

function LoadMimeTypeExtensionList(var MTEList: TMimeTypeExtensionList): Cardinal;
begin
  if MTEList = NIL then MTEList := TMimeTypeExtensionList.Create;
  if MTEList <> NIL then with MTEList do begin
    CaseSensitive := FALSE;
    Duplicates := dupIgnore;
    Sorted := TRUE;
    Result := GetRegKeyList(HKEY_CLASSES_ROOT, 'Mime\Database\Content Type', MTEList, 'Extension', TextAndMediaOnly);
    Add(MIME_PNG + NameValueSeparator + PNG);
    Add(MIME_SVG + NameValueSeparator + SVG);
    Add(MIME_WEBP + NameValueSeparator + WEBP);
  end
  else Result := 0;
end;

//
// Flags for FindMimeFromData
//
const
  FMFD_DEFAULT						= 00000000;
  FMFD_URLASFILENAME					= 00000001;
  FMFD_ENABLEMIMESNIFFING				= 00000002;
  FMFD_IGNOREMIMETEXTPLAIN				= 00000004;
  FMFD_SERVERMIME					= 00000008;
  FMFD_RESPECTTEXTPLAIN					= 00000010;
  FMFD_RETURNUPDATEDIMGMIMES				= 00000020;

function GetMimeTypeExtension(const MimeType: String): String;
begin
  Result := MimeTypeExtensionList.GetMimeTypeExtensionW(E, MimeType);
end;

function ExtractMimeTypeExtension(const FileName: String): String;
begin
  Result := MimeTypeExtensionList.GetMimeTypeExtensionW(FileName, GetFileMimeType(FileName));
end;

function ValidateFileExtension(const FileNameOrExtension: String): String;
begin
  with MimeTypeExtensionList do Result := GetFileTypeByMimeTypeW(GetMimeTypeByExtensionW(FileNameOrExtension));
end;

function IsFileExtSvgXml(const FileNameExt1, FileNameExt2: String): Longbool;
begin
  Result := StrTools.SameText(FileNameExt1, SVG) and StrTools.SameText(FileNameExt2, XML) or
	    StrTools.SameText(FileNameExt1, XML) and StrTools.SameText(FileNameExt2, SVG);
end;

function GetDataMimeType(Data: Pointer; Size: Cardinal; const MimeProposed: String = E): String;
var
  wcMimeType: PChar;
begin
  if FindMimeFromData(NIL, NIL, Data, Size, PChar(MimeProposed),
		      FMFD_ENABLEMIMESNIFFING or FMFD_IGNOREMIMETEXTPLAIN or FMFD_RETURNUPDATEDIMGMIMES,
		      wcMimeType, 0) = S_OK then begin
    try
      Result := wcMimeType;
    finally
      CoTaskMemFree(wcMimeType);
    end;
  end;
end;

function GetFileMimeType(const FileName: String; const DefaultType: String = E; const FileData: RawByteString = E): String;

  function CheckDocType(MimeType: PChar; const DefType: String; const FileData: RawByteString): String;
  var
    sDocType: String;
  begin
    try
      Result := MimeType;
      if StrTools.SameText(MIME_XML, Result) then begin
	sDocType := GetDocType(FileData);
	if StrTools.SameText(DefType, MIME_SVG) and StrTools.SameText(sDocType, DOC_SVG) then
	  Result := DefType
	else if StrTools.SameText(sDocType, DOC_HTML) then Result := MIME_HTML;
      end;
    finally
      CoTaskMemFree(MimeType);
    end;
  end;

var
  hFile: THandle;
  dwSize, dwResult: DWord;
  rsFileData: RawByteString;
  sDefaultType: String;
  wcMimeType: PChar;
begin
  if DefaultType = E then
    sDefaultType := MIME_TEXT
  else sDefaultType := DefaultType;
  if FileName <> E then begin
    hFile := CreateFile(PChar(FileName), GENERIC_READ, FILE_SHARE_READ, NIL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, 0);
    if hFile <> INVALID_HANDLE_VALUE then begin
      try
	dwSize := PEEK_SIZE;
	SetString(rsFileData, NIL, dwSize);
	if ReadFile(hFile, Pointer(rsFileData)^, dwSize, dwResult, NIL) and (dwResult > 0)
	and (FindMimeFromData(NIL, PChar(FileName), Pointer(rsFileData), dwResult, NIL,
			      FMFD_URLASFILENAME or FMFD_ENABLEMIMESNIFFING or
			      FMFD_IGNOREMIMETEXTPLAIN or FMFD_RETURNUPDATEDIMGMIMES,
			      wcMimeType, 0) = S_OK) then begin
	  Result := CheckDocType(wcMimeType, DefaultType, rsFileData);
	  Exit;
	end;
      finally
	CloseHandle(hFile);
      end
    end
  end;
  if FindMimeFromData(NIL, PChar(FileName), Pointer(FileData), Length(FileData), NIL,
		      FMFD_ENABLEMIMESNIFFING or FMFD_IGNOREMIMETEXTPLAIN or FMFD_RETURNUPDATEDIMGMIMES,
		      wcMimeType, 0) = S_OK then begin
    Result := CheckDocType(wcMimeType, DefaultType, FileData);
    Exit;
  end;
  Result := E;
end;

function IsApplicationType(const FileName: String; const DefaultType: String = E; const FileData: RawByteString = E): Boolean;
var
  sMimeType: String;
begin
  if DefaultType = E then
    sMimeType := MIME_TEXT
  else sMimeType := DefaultType;
  Result := StrTools.StartsText(MIME_APP, GetFileMimeType(FileName, sMimeType, FileData));
end;

function IsImageType(const FileName: String; const DefaultType: String = E; const FileData: RawByteString = E): Boolean;
var
  sMimeType: String;
begin
  if DefaultType = E then
    sMimeType := MIME_TEXT
  else sMimeType := DefaultType;
  Result := StrTools.StartsText(MIME_IMG, GetFileMimeType(FileName, sMimeType, FileData));
end;

function IsHyperTextType(const FileName: String): Boolean;
var
  sMimeType: String;
begin
  sMimeType := GetFileMimeType(FileName, MIME_TEXT, E);
  Result := (StrTools.SameText(MIME_HTML, sMimeType) or StrTools.SameText(MIME_XML, sMimeType));
end;

function IsSvgType(const FileName: String): Boolean;
var
  sMimeType: String;
begin
  sMimeType := GetFileMimeType(FileName, MIME_SVG, E);
  Result := StrTools.SameText(MIME_SVG, sMimeType) or StrTools.SameText(MIME_XML, sMimeType);
end;

function FindFile(const FileMask: AnsiString; const ExcludeExt: AnsiString = E): AnsiString;
var
  hSearch: THandle;
  fd: TWin32FindDataA;
begin
  Result := E;
  hSearch := FindFirst(PAnsiChar(FileMask), fd);
  try
    while hSearch <> INVALID_HANDLE_VALUE do begin
      if (fd.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY = 0) then begin
	Result := FileTools.ExtractFilePath(FileMask) + fd.cFileName;
	if (Result <> E) and ((ExcludeExt = E) or not SameFileExt(Result, ExcludeExt)) then Break;
      end;
      if not FindNext(hSearch, fd) then Break;
    end;
  finally
    Windows.FindClose(hSearch);
  end;
end;

function FindFile(const FileMask: UnicodeString; const ExcludeExt: UnicodeString = E): UnicodeString;
var
  hSearch: THandle;
  fd: TWin32FindDataW;
begin
  Result := E;
  hSearch := FindFirst(PWideChar(FileMask), fd);
  try
    while hSearch <> INVALID_HANDLE_VALUE do begin
      if (fd.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY = 0) then begin
	Result := FileTools.ExtractFilePath(FileMask) + fd.cFileName;
	if (Result <> E) and ((ExcludeExt = E) or not SameFileExt(Result, ExcludeExt)) then Break;
      end;
      if not FindNext(hSearch, fd) then Break;
    end;
  finally
    Windows.FindClose(hSearch);
  end;
end;

function FindFirst(const FileMask: AnsiString; var FindData: TWin32FindDataA): THandle; inline;
begin
  Result := FindFirstFileA(PAnsiChar(FileMask), FindData);
end;

function FindFirst(const FileMask: UnicodeString; var FindData: TWin32FindDataW): THandle; inline;
begin
  Result := FindFirstFileW(PWideChar(FileMask), FindData);
end;

function FindNext(Search: THandle; var FindData: TWin32FindDataA): Longbool; inline;
begin
  Result := FindNextFileA(Search, FindData);
end;

function FindNext(Search: THandle; var FindData: TWin32FindDataW): Longbool; inline;
begin
  Result := FindNextFileW(Search, FindData);
end;

function FindClose(Search: THandle): Longbool; inline;
begin
  Result := Windows.FindClose(Search);
end;

function DeleteFiles(const FileMask: AnsiString): Longbool;
var
  sPath: AnsiString;
  hSearch: THandle;
  fd: TWin32FindDataA;
begin
  Result := false;
  sPath := FileTools.ExtractFilePath(FileMask);
  hSearch := FindFirst(FileMask, fd);
  if hSearch <> INVALID_HANDLE_VALUE then begin
    try
      repeat
	if (fd.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) = 0 then begin
	  Result := DeleteFile(PAnsiChar(sPath + fd.cFileName));
	end;
      until not FindNext(hSearch, fd);
    finally
      Windows.FindClose(hSearch);
    end;
  end;
end;

function DeleteFiles(const FileMask: UnicodeString): Longbool;
var
  sPath: UnicodeString;
  hSearch: THandle;
  fd: TWin32FindDataW;
begin
  Result := false;
  sPath := FileTools.ExtractFilePath(FileMask);
  hSearch := FindFirst(FileMask, fd);
  if hSearch <> INVALID_HANDLE_VALUE then begin
    try
      repeat
	if (fd.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) = 0 then begin
	  Result := DeleteFile(PWideChar(sPath + fd.cFileName));
	end;
      until not FindNext(hSearch, fd);
    finally
      Windows.FindClose(hSearch);
    end;
  end;
end;

function ExistsLockedOrShared(const Filename: RawByteString): Boolean; overload;
var
  FindData: TWin32FindDataA;
  LHandle: THandle;
begin
  if not StartsDSorBS(Filename) then begin
    LHandle := FindFirstFileA(PAnsiChar(Filename), FindData);
    if LHandle <> INVALID_HANDLE_VALUE then begin
      Windows.FindClose(LHandle);
      Result := (FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY = 0);
      Exit;
    end;
  end;
  Result := FALSE;
end;

function ExistsLockedOrShared(const Filename: UnicodeString): Boolean; overload;
var
  FindData: TWin32FindDataW;
  LHandle: THandle;
begin
  if not StartsDSorBS(Filename) then begin
    LHandle := FindFirstFileW(PWideChar(Filename), FindData);
    if LHandle <> INVALID_HANDLE_VALUE then begin
      Windows.FindClose(LHandle);
      Result := (FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY = 0);
      Exit;
    end;
  end;
  Result := False;
end;

function DeleteFile(const Name: RawByteString; IncludeReadOnly: Longbool = FALSE): Longbool;
begin
  if IncludeReadOnly then SetFileAttributesA(Pointer(Name), GetFileAttributesA(Pointer(Name)) and not FILE_ATTRIBUTE_READONLY);
  Result := Windows.DeleteFileA(PAnsiChar(Name));
end;

function DeleteFile(const Name: UnicodeString; IncludeReadOnly: Longbool = FALSE): Longbool;
begin
  if IncludeReadOnly then SetFileAttributesW(Pointer(Name), GetFileAttributesW(Pointer(Name)) and not FILE_ATTRIBUTE_READONLY);
  Result := Windows.DeleteFileW(PWideChar(Name));
end;

function StartsDSorBS(const Name: RawByteString): Longbool;
const
  DBS = DWord(Ord(SL) SHL 8 OR Ord(SL));
  DBBS = DWord(Ord(BS) SHL 8 OR Ord(BS));
asm
  {$IFDEF CPUX64}
  // RCX = Pointer(Name)
  OR	RCX,RCX
  JZ	@FALSE
  MOV	RCX,[RCX]
  CMP	CX,DBS
  JZ   @TRUE
  CMP	CX,DBBS
  JNZ	@FALSE
  @TRUE:
  XOR	RAX,RAX
  DEC	RAX
  RET
  @FALSE:
  XOR	EAX,EAX
  @EXIT:
  RET
  {$ELSE}
  // EAX = Pointer(Name)
  OR	EAX,EAX
  JZ	@EXIT
  MOV	EAX,[EAX]
  CMP	AX,DBS
  JZ   @TRUE
  CMP	AX,DBBS
  JNZ	@FALSE
  @TRUE:
  XOR	EAX,EAX
  DEC	EAX
  RET
  @FALSE:
  XOR	EAX,EAX
  @EXIT:
  RET
  {$ENDIF}
end;

function StartsDSorBS(const Name: UnicodeString): Longbool;
const
  DBS = DWord(Ord(SL) SHL 16 OR Ord(SL));
  DBBS = DWord(Ord(BS) SHL 16 OR Ord(BS));
asm
  {$IFDEF CPUX64}
  // RCX = Pointer(Name)
  OR	RCX,RCX
  JZ	@FALSE
  MOV	RCX,[RCX]
  CMP 	ECX,DBS
  JZ   @TRUE
  CMP	ECX,DBBS
  JNZ	@FALSE
  @TRUE:
  XOR	RAX,RAX
  DEC	RAX
  @FALSE:
  XOR	EAX,EAX
  @EXIT:
  RET
  {$ELSE}
// EAX = Pointer(Name)
  OR	EAX,EAX
  JZ	@EXIT
  MOV	EAX,[EAX]
  CMP 	EAX,DBS
  JZ   @TRUE
  CMP	EAX,DBBS
  JNZ	@FALSE
  @TRUE:
  XOR	EAX,EAX
  DEC	EAX
  RET
  @FALSE:
  XOR	EAX,EAX
  @EXIT:
  RET
  {$ENDIF}
end;


function FileExists(const Name: RawByteString; IncludeRemote: Longbool = FALSE): Longbool;
var
  Code, LastError: Cardinal;
begin
  if not StartsDSorBS(Name) then begin
    Code := GetFileAttributesA(PAnsiChar(Name));
    if Code <> INVALID_FILE_ATTRIBUTES then
      Result := (FILE_ATTRIBUTE_DIRECTORY and Code = 0)
    else begin
      LastError := GetLastError;
      Result := (LastError <> ERROR_FILE_NOT_FOUND) and (LastError <> ERROR_PATH_NOT_FOUND) and
		(LastError <> ERROR_INVALID_NAME) and ExistsLockedOrShared(Name);
    end;
    Exit;
  end
  else if IncludeRemote then begin
    Result := PathFileExistsA(PAnsiChar(Name));
    Beep;
    Exit;
  end;
  Result := FALSE;
end;

function FileExists(const Name: UnicodeString; IncludeRemote: Longbool = FALSE): Longbool;
var
  Code, LastError: Cardinal;
begin
  if not StartsDSorBS(Name) then begin
    Code := GetFileAttributesW(PWideChar(Name));
    if Code <> INVALID_FILE_ATTRIBUTES then
      Result := (FILE_ATTRIBUTE_DIRECTORY and Code = 0)
    else begin
      LastError := GetLastError;
      Result := (LastError <> ERROR_FILE_NOT_FOUND) and (LastError <> ERROR_PATH_NOT_FOUND) and
		(LastError <> ERROR_INVALID_NAME) and ExistsLockedOrShared(Name);
    end;
    Exit;
  end
  else if IncludeRemote then begin
    Result := PathFileExistsW(PWideChar(Name));
    Exit;
  end;
  Result := FALSE;
end;

function DirExists(const Name: RawByteString; LastBackslashPos: Integer = 0): Longbool;
var
  saName: AnsiString;
  cCode: Cardinal;
begin
  if not StartsDSorBS(Name) then begin
    if LastBackslashPos = -1 then LastBackslashPos := LastDelimiter(Name);
    if LastBackslashPos > 0 then
      saName := Copy(Name, 1, pred(LastBackslashPos))
    else saName := Name;
    cCode := GetFileAttributesA(PAnsiChar(saName));
    Result := (cCode <> INVALID_FILE_ATTRIBUTES) and (FILE_ATTRIBUTE_DIRECTORY and cCode <> 0);
    Exit;
  end;
  Result := FALSE;
end;

function DirExists(const Name: UnicodeString; LastBackslashPos: Integer = 0): Longbool;
var
  suName: UnicodeString;
  cCode: Cardinal;
begin
  if not StartsDSorBS(Name) then begin
    if LastBackslashPos = -1 then LastBackslashPos := LastDelimiter(Name);
    if LastBackslashPos > 0 then
      suName := Copy(Name, 1, pred(LastBackslashPos))
    else suName := Name;
    cCode := GetFileAttributesW(PWideChar(suName));
    Result := (cCode <> INVALID_FILE_ATTRIBUTES) and (FILE_ATTRIBUTE_DIRECTORY and cCode <> 0);
    Exit;
  end;
  Result := FALSE;
end;


var
  OldCallWndRetProc: HHOOK = 0;

function NewCallWndRetProc(nCode: Integer; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
begin
  with PCWPRetStruct(lParam)^ do begin
    if (nCode = HC_ACTION) and (Message = WM_SHOWWINDOW) and (wParam <> 0) and IsDialogWindow(hWnd) then begin
      CenterDialog(hWnd, GetWindowLongPtr(hWnd, GWLP_HWNDPARENT));
    end;
  end;
  Result := CallNextHookEx(OldCallWndRetProc, nCode, wParam, lParam);
end;

function FindDir(Owner: THandle; const SearchPath, Prompt: String; AllowCreation: Longbool = FALSE; EnableEdit: Longbool = FALSE): String;

  function BrowseCallbackProc(hwnd: HWND; uMsg: UINT; lParam: LPARAM; lpData: LPARAM): Integer; stdcall;
  begin
    Result := 0;
    case uMsg of
      BFFM_INITIALIZED: if lpData <> 0 then begin
	SendMessage(hwnd, BFFM_SETSELECTION, 1, lpData);
	SendMessage(hwnd, BFFM_SETEXPANDED, 1, lpData);
      end;
    end;
  end;

var
  sPath: String;
  pa: TPathArray;
  bi: TBrowseInfo;
  pidl: PItemIDList;
begin
  sPath := FileTools.ExtractFilePath(SearchPath);
  FillChar(pa, SizeOf(TPathArray), 0);
  FillChar(bi, SizeOf(TBrowseInfo), 0);
  pidl := NIL;
  try
    with bi do begin
      hwndOwner := Owner;
      pszDisplayName := pa;
      lpszTitle := Pointer(Prompt);
      ulFlags := Ord(EnableEdit = TRUE)*BIF_EDITBOX or BIF_NEWDIALOGSTYLE or
		 Ord(AllowCreation = FALSE)*BIF_NONEWFOLDERBUTTON or
		 BIF_RETURNONLYFSDIRS or BIF_UAHINT or BIF_VALIDATE;
      lParam := Windows.LPARAM(sPath);
      lpfn := @BrowseCallbackProc;
    end;
    OldCallWndRetProc := SetWindowsHookEx(WH_CALLWNDPROCRET, @NewCallWndRetProc, 0, GetCurrentThreadID);
    try
      pidl := SHBrowseForFolder(bi);
      ProcessMessages;
    finally
      UnhookWindowsHookEx(OldCallWndRetProc);
      OldCallWndRetProc := 0;
    end;
    Result := GetSystemFolderPath(pidl);
  finally
    if pidl <> NIL then CoTaskMemFree(pidl);
  end;
end;

function FindExecutable(const FileExtension: String): String;
var
  dwSize: DWord;
begin
  dwSize := 0;
  if AssocQueryString(0, ASSOCSTR_EXECUTABLE, PChar(FileExtension), NIL, NIL, @dwSize) = S_FALSE then begin
    SetString(Result, NIL, pred(dwSize));
    if AssocQueryString(0, ASSOCSTR_EXECUTABLE, PChar(FileExtension), NIL, Pointer(Result), @dwSize) = S_OK then Exit;
  end;
  Result := E;
end;

function ReloadModule(const FileName: String): THandle;
begin
  Result := LoadLibrary(PChar(GetModuleName(GetModuleHandle(PChar(FileName)))));
end;

function LoadResModule(const FileName: String): THandle;
begin
  Result := LoadLibraryEx(PChar(FileName), LOAD_LIBRARY_AS_DATAFILE);
end;

function ReloadAsResModule(Module: THandle): THandle;
begin
  Result := LoadLibraryEx(PChar(GetModuleName(Module)), LOAD_LIBRARY_AS_DATAFILE);
end;

function DialogItemInt(Dialog: THandle; ItemID: Cardinal): Integer;
var
  fSuccess: Longbool;
begin
  Result := GetDlgItemInt(Dialog, ItemID, fSuccess, TRUE);
end;

function DialogItemString(Dialog: THandle; ItemID: Cardinal): String;
begin
  SetString(Result, NIL, $FFFF);
  Delete(Result, succ(GetDlgItemText(Dialog, ItemID, PChar(Result), $FFFF)), MAXINT);
end;

type
  TDialogItemParams = Record
    Extended: Longbool;
    Offset: Pointer;
    Count: Integer;
  end;

function GetNextPWideChar(var p: PAnsiChar): PChar;
var
  pw: PWideChar;
begin
  pw := PWideChar(p);
  case pw^ of
    WideChar(0): inc(pw);
    WideChar($FFFF): inc(pw, 2);
  else
    Result := pw;
    inc(pw, succ(StrTools.StrLen(pw)));
    p := PAnsiChar(pw);
    Exit;
  end;
  p := PAnsiChar(pw);
  Result := E;
end;

type
  TTypeFace = Array[0..0] of WideChar;
  PFontDef = ^TFontDef;
  TFontDef = Record
    pointsize: Word;
    typeface: TTypeFace;
  end;
  PFontDefEx = ^TFontDefEx;
  TFontDefEx = Packed Record
    pointsize: Word;
    weight: Word;
    italic: Byte;
    charset: Byte;
    typeface: TTypeFace;
  end;
  PDlgTemplate = ^TDlgTemplate;
  TDlgTemplate = Packed Record
    style: DWord;
    exStyle: DWord;
    cDlgItems: Word;
    x: Short;
    y: Short;
    cx: Short;
    cy: Short;
  end;
  PDlgTemplateEx = ^TDlgTemplateEx;
  TDlgTemplateEx = Packed Record
    dlgVer: Word;
    signature: Word;
    helpID: DWord;
    exStyle: DWord;
    style: DWord;
    cDlgItems: Word;
    x: Short;
    y: Short;
    cx: Short;
    cy: Short;
  end;

function GetDialogCaption(Module: THandle; DialogID: PChar; var DialogItemParams: TDialogItemParams): String;
var
  pDialog: PAnsiChar;
  fFontDef: Longbool;
begin
  pDialog := LockResource(LoadResource(Module, FindResource(Module, DialogID, RT_DIALOG)));
  if pDialog <> NIL then begin
    DialogItemParams.Extended := PDlgTemplateEx(pDialog)^.signature = $FFFF;
    DialogItemParams.Offset := NIL;
    if DialogItemParams.Extended then with PDlgTemplateEx(pDialog)^ do begin
      fFontDef := (style and DS_SETFONT <> 0);
      DialogItemParams.Count := cDlgItems;
      inc(pDialog, SizeOf(TDlgTemplateEx));
    end
    else with PDlgTemplate(pDialog)^ do begin
      fFontDef := (style and DS_SETFONT <> 0);
      DialogItemParams.Count := cDlgItems;
      inc(pDialog, SizeOf(TDlgTemplate));
    end;
    // menu
    GetNextPWideChar(pDialog);
    // class name
    GetNextPWideChar(pDialog);
    // dialog caption
    Result := GetNextPWideChar(pDialog);
    if fFontDef then begin
      // numeric font params
      if DialogItemParams.Extended then
	inc(pDialog, SizeOf(TFontDefEx) - SizeOf(TTypeFace))
      else inc(pDialog, SizeOf(TFontDef) - SizeOf(TTypeFace));
      // font type face name
      GetNextPWideChar(pDialog);
    end;
    with DialogItemParams do if Count > 0 then Offset := pDialog;
    Exit;
  end;
  FillChar(DialogItemParams, SizeOf(TDialogItemParams), 0);
  Result := E;
end;

type
  TExtraCount = Word;
  PDlgItemTemplate = ^TDlgItemTemplate;
  TDlgItemTemplate = Packed Record
    style: DWord;
    exStyle: DWord;
    x: Short;
    y: Short;
    cx: Short;
    cy: Short;
    id: Word;
  end;
  PDlgItemTemplateEx = ^TDlgItemTemplateEx;
  TDlgItemTemplateEx = Packed Record
    helpID: DWord;
    exStyle: DWord;
    style: DWord;
    x: Short;
    y: Short;
    cx: Short;
    cy: Short;
    id: DWord;
  end;

function GetDialogControlText(const DialogItemParams: TDialogItemParams; ItemID: Integer): String;
var
  pItem: PAnsiChar;
  iIndex, iItemID, iExtraCount: Integer;
begin
  with DialogItemParams do if Count > 0 then begin
    pItem := Offset;
    for iIndex := 1 to Count do begin
      inc(pItem, (4 - (DWord(pItem) and 3)) and 3);
      if Extended then begin
	iItemID := PDlgItemTemplateEx(pItem)^.id;
	inc(pItem, SizeOf(TDlgItemTemplateEx));
      end
      else begin
	iItemID := PDlgItemTemplate(pItem)^.id;
	inc(pItem, SizeOf(TDlgItemTemplate));
      end;
      // control class
      GetNextPWideChar(pItem);
      // control text
      if iItemID = ItemID then begin
	Result := GetNextPWideChar(pItem);
	Exit;
      end
      else GetNextPWideChar(pItem);
      // extra count
      iExtraCount := Word(pItem^);
      inc(pItem, SizeOf(TExtraCount));
      // creation data
      if iExtraCount > 0 then begin
	if Extended then
	  inc(pItem, iExtraCount)
	else inc(pItem, iExtraCount + SizeOf(TExtraCount));
      end
    end;
  end;
  Result := E;
end;

function DialogFontInfo(Module: THandle; DialogID: PChar): TDlgFont;
var
  pDialog: PAnsiChar;
  dip: TDialogItemParams;
  fFontDef: Longbool;
begin
  pDialog := LockResource(LoadResource(Module, FindResource(Module, DialogID, RT_DIALOG)));
  if pDialog <> NIL then begin
    dip.Extended := PDlgTemplateEx(pDialog)^.signature = $FFFF;
    dip.Offset := NIL;
    if dip.Extended then with PDlgTemplateEx(pDialog)^ do begin
      fFontDef := (style and DS_SETFONT <> 0);
      dip.Count := cDlgItems;
      inc(pDialog, SizeOf(TDlgTemplateEx));
    end
    else with PDlgTemplate(pDialog)^ do begin
      fFontDef := (style and DS_SETFONT <> 0);
      dip.Count := cDlgItems;
      inc(pDialog, SizeOf(TDlgTemplate));
    end;
    // menu
    GetNextPWideChar(pDialog);
    // class name
    GetNextPWideChar(pDialog);
    // dialog caption
    GetNextPWideChar(pDialog);
    if fFontDef then begin
      // numeric font params
      Result.Size := PFontDef(pDialog)^.pointsize;
      if dip.Extended then
	inc(pDialog, SizeOf(TFontDefEx) - SizeOf(TTypeFace))
      else inc(pDialog, SizeOf(TFontDef) - SizeOf(TTypeFace));
      // font type face name
      Result.Name := GetNextPWideChar(pDialog);
      Exit;
    end;
  end;
  with Result do begin
    Name := E;
    Size := 0;
  end;
end;

function DialogCaptionString(Module: THandle; DialogID: PChar): String;
var
  dip: TDialogItemParams;
begin
  Result := GetDialogCaption(Module, DialogID, dip);
end;

function DialogControlString(Module: THandle; DialogID: PChar; ControlID: Cardinal): String;
var
  dip: TDialogItemParams;
begin
  GetDialogCaption(Module, DialogID, dip);
  Result := GetDialogControlText(dip, ControlID);
end;

function MenuItemString(Module: THandle; MenuID: PChar; ItemID: Cardinal): String;
type
  TMenuString = Array[0..0] of WideChar;
  PMenuItemTemplateHeader = ^TMenuItemTemplateHeader;
  TMenuItemTemplateHeader = packed record
    versionNumber: Word;
    offset: Word;
  end;
  PMenuItemTemplateHeaderEx = ^TMenuItemTemplateHeaderEx;
  TMenuItemTemplateHeaderEx = packed record
    versionNumber: Word;
    offset: Word;
    dwHelpId: DWord;
  end;
  PMenuItemTemplate = ^TMenuItemTemplate;
  TMenuItemTemplate = packed record
    mtOption: Word;
    mtID: Word;
    mtString: TMenuString;
  end;
  PMenuItemTemplateEx = ^TMenuItemTemplateEx;
  TMenuItemTemplateEx = packed record
    dwHelpId: DWord;
    dwType: DWord;
    dwState: DWord;
    menuId: DWord;
    bResInfo: Word;
    szText: TMenuString;
  end;
var
  hResource: THandle;
  dwResourceSize: DWord;
  pMenu, pOffset: PAnsiChar;
  fExtended: Longbool;
begin
  hResource := FindResource(Module, MenuID, RT_MENU);
  dwResourceSize := SizeOfResource(Module, hResource);
  pMenu := LockResource(LoadResource(Module, hResource));
  if pMenu <> NIL then begin
    fExtended := (PMenuItemTemplateHeader(pMenu)^.versionNumber = 1);
    pOffset := Ptr(DWord(pMenu) + SizeOf(TMenuItemTemplateHeader) + PMenuItemTemplateHeader(pMenu)^.offset);
    if fExtended then begin
      while DWord(pOffset - pMenu) < dwResourceSize do with PMenuItemTemplateEx(pOffset)^ do begin
	if menuID = ItemID then begin
	  Result := szText;
	  Exit;
	end
	else inc(pOffset, SizeOf(TMenuItemTemplateEx) + StrTools.StrLen(szText) shl 1);
      end;
    end
    else begin
      while DWord(pOffset - pMenu) < dwResourceSize do with PMenuItemTemplate(pOffset)^ do begin
	if (mtOption and MF_POPUP) <> 0 then
	  inc(pOffset, SizeOf(TMenuItemTemplate) + StrTools.StrLen(mtString) shl 1)
	else begin
	  if mtID = ItemID then begin
	    SetString(Result, mtString, StrTools.StrLen(mtString));
	    Exit;
	  end
	  else inc(pOffset, (SizeOf(TMenuItemTemplate) + StrTools.StrLen(mtString) shl 1));
	end;
      end;
    end;
  end;
end;

function LoadHtmlResource(Module: THandle; const ResourceName: String; var Resource: TResource): Longbool;
begin
  with Resource do begin
    Handle := FindResource(Module, PChar(ResourceName), RT_HTML);
    if Handle <> 0 then begin
      Data.Handle := LoadResource(Module, Handle);
      Exit;
    end;
  end;
  FillChar(Result, SizeOf(Result), 0)
end;

function LoadHtmlString(Module: THandle; const Name: String): RawByteString;
var
  rHtml: TResource;
begin
  if LoadHtmlResource(Module, Name, rHtml) then with rHtml do begin
    if Data.Handle <> 0 then begin
      SetString(Result, PAnsiChar(Data), SizeofResource(Module, Handle));
      Exit;
    end;
  end;
  Result := E;
end;


initialization
  if MimeTypeExtensionList = NIL then LoadMimeTypeExtensionList(MimeTypeExtensionList);

finalization
  if MimeTypeExtensionList <> NIL then FreeAndNil(MimeTypeExtensionList);

end.

