{------------------------------------------------------------------------------}
{        This code is copyright (C) 2008-2025 by Michael in der Wiesche        }
{------------------------------------------------------------------------------}
{$J+,Z4}
unit MDResources;

interface

uses
  Windows, Messages, SysUtils, Processes, StrUtils, StrTools, FileTools, PMMessages, PMOptions, PMTools;

{$I *.INC}

const
  CFG = '.cfg';
  DLL = '.dll';
  DMP = '.dmp';
  FFF = '.fff';
  TXT = '.txt';

const
  MD_BASE_NAME = 'MiniDump';
  MD_LIBRARY = MD_BASE_NAME + DLL;
  MD_FORMFACTS = MD_BASE_NAME + FFF;
  MD_DOWNLOAD_URL = 'https://www.pmpgp.de/pmtools/MiniDumpAndSources.zip';

type
  TLangIndex = (liEN, liDE, liFR, liIT);
  TLangIDs = Record
    Index: TLangIndex;
    Offset: Integer;
    Module: THandle;
  end;

function Str(ID: Integer): String;
function GetBaseDTFileName: String;
function GetEnvironmentInfo: String;
function GetVersionInformation: String;
function GetMiniDumpInstallPath: String;
function GetLangOffset: Integer;
function GetLangModule: THandle;
procedure GetLangIndex(const LibName: String; PMFormApi: TPMFormApi);
procedure SetLangIndex(const LangTLD: String; ModuleHandle: THandle);
function GetHelpStrings(var HelpText, HelpCaption: String): Longbool;
function UpdateAvailable(hWnd: THandle; LangOffset: Integer): String;
function SetControlText(hWnd: THandle; LangOffset, ControlID: Integer; const Text: String): Longbool;

implementation

const
  MDInstance: THandle = 0;
  LangIDs: TLangIDs = (Index: liEN; Offset: 0; Module: 0);
  LangTLDs: Array[TLangIndex] of String = ('', 'DE', 'FR', 'IT');

function Str(ID: Integer): String;
begin
  Result := ModuleString(MDInstance, ID);
end;

function GetBaseDTFileName: String;
var
  st: TSystemTime;
begin
  GetSystemTime(st);
  Result := DateString(st, 'yyyy-MM-dd') + HT + TimeString(st, 'HH-mm-ss');
end;

function GetEnvironmentInfo: String;
begin
  Result := WinVersion(FALSE, FALSE, FALSE, TRUE);
end;

function GetVersionInformation: String;
var
  fvi: TFileVersionInfo;
begin
  with fvi do begin
    QueryFile := GetModuleName(HInstance);
    QueryParameters := [fviProductVersion];
    if FileVersionInfo(fvi) then begin
      Result := LB + ExtractBaseName(QueryFile) + SP + ProductVersion + RB;
      Exit;
    end;
  end;
  Result := E;
end;

function GetMiniDumpInstallPath: String;
const
  REGKEY_UNINSTALL = 'Software\Microsoft\Windows\CurrentVersion\Uninstall';
  REGKEY_MINIDUMP = '{CEF609C4-3E78-41EB-BC61-582EC4BE2086}_is1';
  REGSTR_INSTALL = 'InstallLocation';
begin
  Result := AddBackslash(GetRegString(HKEY_LOCAL_MACHINE, REGKEY_UNINSTALL + BS + REGKEY_MINIDUMP, REGSTR_INSTALL)) + AddBackslash(MD_BASE_NAME);
end;

function GetProductVersion: String;
var
  fvi: TFileVersionInfo;
begin
  with fvi do begin
    QueryFile := GetModuleName(HInstance);
    QueryParameters := [fviProductVersion];
    if FileVersionInfo(fvi) then begin
      Result := ProductVersion;
      Exit;
    end;
  end;
  Result := E;
end;

function GetLangOffset: Integer;
begin
  Result := LangIDs.Offset;
end;

function GetLangModule: THandle;
begin
  Result := LangIDs.Module;
end;

procedure GetLangIndex(const LibName: String; PMFormApi: TPMFormApi);
const
  WPM32 = 'WPM32';
  WPM32_ = WPM32 + UD;
var
  iLIDPos: Integer;
  sModule: String;
begin
  sModule := FileTools.ExtractFileName(LibName);
  if StrTools.StartsText(WPM32_, sModule) then
    iLIDPos := succ(Length(WPM32_))
  else if StrTools.StartsText(WPM32, sModule) then
    iLIDPos := succ(Length(WPM32))
  else Exit;
  SetLangIndex(Copy(sModule, iLIDPos, 2), GetModuleHandle(PChar(sModule)));
end;

procedure SetLangIndex(const LangTLD: String; ModuleHandle: THandle);
begin
  with LangIDs do begin
    if StrTools.SameText(LangTLD, 'EN') then
      Index := liEN
    else Index := TLangIndex(StrTools.IndexText(LangTLD, LangTLDs));
    Offset := Ord(Index) * OFFSET_LANGUAGE;
    if Module = 0 then Module := ReloadAsResModule(ModuleHandle);
  end;
end;

function GetHelpStrings(var HelpText, HelpCaption: String): Longbool;
const
  RM = 'MDReadMe';
  VER = 'VERSION';
var
  sHelpFile: String;
  iBeg, iEnd: Integer;
begin
  Result := FALSE;
  if LangIDs.Index = liEN then
    sHelpFile := FileTools.ExtractFilePath(GetModuleName(HInstance)) + RM + TXT
  else sHelpFile := FileTools.ExtractFilePath(GetModuleName(HInstance)) + RM + LangTLDs[LangIDs.Index] + TXT;
  HelpCaption := String(ReadTextFile(sHelpFile));
  if HelpCaption <> E then begin
    HelpText := ExtractStr(HelpCaption, FirstStrPos(PARA, HelpCaption) + Length(PARA), MAXINT);
    Delete(HelpCaption, FirstStrPos(CRLF, HelpCaption), MAXINT);
    Delete(HelpText, LastStrPos(PARA + VER, HelpText), MAXINT);
    iBeg := FirstCharPos(LB, HelpCaption);
    iEnd := NextCharPos(RB, HelpCaption, iBeg);
    Delete(HelpCaption, iBeg, iEnd - iBeg + 2);
    Result := TRUE;
  end
  else HelpText := SysErrorMessage(ERROR_READ_FAULT);
end;

function UpdateAvailable(hWnd: THandle; LangOffset: Integer): String;
begin
  Result := Str(LangOffset + STR_UPDATE_LOCAL_ERROR);
  if ExtensionUpdateFound(NIL, MD_DOWNLOAD_URL, Result) then
    Result := Str(LangOffset + STR_UPDATE_STATUS)
  else if Result = E then
    Result := Str(LangOffset + STR_UPDATE_CURRENT)
  else if Result <> Str(LangOffset + STR_UPDATE_LOCAL_ERROR) then Result := Str(LangOffset + STR_UPDATE_REMOTE_ERROR) + CRLF + Result;
end;

function SetControlText(hWnd: THandle; LangOffset, ControlID: Integer; const Text: String): Longbool;
var
  sText: String;
begin
  if Text = E then begin
    sText := Str(LangOffset + ControlID);
    if sText = E then Str(ControlID);
  end
  else sText := Text;
  Result := SetDlgItemText(hWnd, ControlID, PChar(sText));
end;

procedure LoadMiniDump;
begin
  if FileTools.ExtractFileName(GetModuleName(HInstance)) <> MD_LIBRARY then
    MDInstance := LoadResModule(GetMiniDumpInstallPath + MD_LIBRARY)
  else MDInstance := hInstance;
end;

begin
  LoadMiniDump;
end.

